import contextlib
import dataclasses
import enum
import io
from typing import Generic, List, NewType, Tuple, Type, TypeVar, Union

import pytest
import yaml
from typing_extensions import Annotated

import tyro

T = TypeVar("T")


def _check_serialization_identity(cls: Type[T], instance: T) -> None:
    assert tyro.extras.from_yaml(cls, tyro.extras.to_yaml(instance)) == instance


ScalarType = TypeVar("ScalarType")


def test_tuple_generic_variable() -> None:
    @dataclasses.dataclass
    class TupleGenericVariable(Generic[ScalarType]):
        xyz: Tuple[ScalarType, ...]

    assert tyro.cli(
        TupleGenericVariable[int], args=["--xyz", "1", "2", "3"]
    ) == TupleGenericVariable((1, 2, 3))


def test_tuple_generic_variable_newtype() -> None:
    @dataclasses.dataclass
    class TupleGenericVariable(Generic[ScalarType]):
        xyz: Tuple[ScalarType, ...]

    SpecialInt = NewType("SpecialInt", int)
    assert tyro.cli(
        TupleGenericVariable[SpecialInt], args=["--xyz", "1", "2", "3"]
    ) == TupleGenericVariable((SpecialInt(1), SpecialInt(2), SpecialInt(3)))


def test_tuple_generic_variable_newtype_container() -> None:
    @dataclasses.dataclass
    class TupleGenericVariable(Generic[ScalarType]):
        xyz: Tuple[ScalarType, ...]

    SpecialInt = NewType("SpecialInt", Tuple[int])
    assert tyro.cli(
        TupleGenericVariable[SpecialInt], args=["--xyz", "1", "2", "3"]
    ) == TupleGenericVariable((SpecialInt((1,)), SpecialInt((2,)), SpecialInt((3,))))


def test_tuple_generic_variable_more_newtype() -> None:
    @dataclasses.dataclass
    class TupleGenericVariable(Generic[ScalarType]):
        xyz: Tuple[ScalarType, ...]

    SpecialInt = NewType("SpecialInt", int)
    SpecialTuple = NewType("SpecialTuple", TupleGenericVariable[SpecialInt])
    assert tyro.cli(SpecialTuple, args=["--xyz", "1", "2", "3"]) == SpecialTuple(
        TupleGenericVariable((SpecialInt(1), SpecialInt(2), SpecialInt(3)))
    )


def test_tuple_generic_helptext() -> None:
    @dataclasses.dataclass
    class TupleGenericVariableHelptext(Generic[ScalarType]):
        """Helptext!"""

        xyz: Tuple[ScalarType, ...]

    f = io.StringIO()
    with pytest.raises(SystemExit):
        with contextlib.redirect_stdout(f):
            tyro.cli(TupleGenericVariableHelptext[int], args=["--help"])
    helptext = f.getvalue()
    assert "Helptext!" in helptext


def test_tuple_generic_no_helptext() -> None:
    @dataclasses.dataclass
    class TupleGenericVariableNoHelptext(Generic[ScalarType]):
        xyz: Tuple[ScalarType, ...]

    f = io.StringIO()
    with pytest.raises(SystemExit):
        with contextlib.redirect_stdout(f):
            tyro.cli(TupleGenericVariableNoHelptext[int], args=["--help"])
    helptext = f.getvalue()
    assert "Helptext!" not in helptext

    # Check that we don't accidentally grab docstrings from the generic alias!
    assert "The central part of internal API" not in helptext


def test_tuple_generic_fixed() -> None:
    @dataclasses.dataclass
    class TupleGenericFixed(Generic[ScalarType]):
        xyz: Tuple[ScalarType, ScalarType, ScalarType]

    assert tyro.cli(
        TupleGenericFixed[int], args=["--xyz", "1", "2", "3"]
    ) == TupleGenericFixed((1, 2, 3))


class CoordinateFrame(enum.Enum):
    WORLD = enum.auto()
    CAMERA = enum.auto()


@dataclasses.dataclass
class Point3(Generic[ScalarType]):
    x: ScalarType
    y: ScalarType
    z: ScalarType
    frame: CoordinateFrame


def test_simple_generic() -> None:
    @dataclasses.dataclass
    class SimpleGeneric:
        point_continuous: Point3[float]
        point_discrete: Point3[int]

    parsed_instance = tyro.cli(
        SimpleGeneric,
        args=[
            "--point-continuous.x",
            "1.2",
            "--point-continuous.y",
            "2.2",
            "--point-continuous.z",
            "3.2",
            "--point-continuous.frame",
            "WORLD",
            "--point-discrete.x",
            "1",
            "--point-discrete.y",
            "2",
            "--point-discrete.z",
            "3",
            "--point-discrete.frame",
            "WORLD",
        ],
    )
    assert parsed_instance == SimpleGeneric(
        Point3(1.2, 2.2, 3.2, CoordinateFrame.WORLD),
        Point3(1, 2, 3, CoordinateFrame.WORLD),
    )
    _check_serialization_identity(SimpleGeneric, parsed_instance)

    with pytest.raises(SystemExit):
        # Accidentally pass in floats instead of ints for discrete
        tyro.cli(
            SimpleGeneric,
            args=[
                "--point-continuous.x",
                "1.2",
                "--point-continuous.y",
                "2.2",
                "--point-continuous.z",
                "3.2",
                "--point-continuous.frame",
                "WORLD",
                "--point-discrete.x",
                "1.5",
                "--point-discrete.y",
                "2.5",
                "--point-discrete.z",
                "3.5",
                "--point-discrete.frame",
                "WORLD",
            ],
        )


def test_multilevel_generic() -> None:
    @dataclasses.dataclass
    class Triangle(Generic[ScalarType]):
        a: Point3[ScalarType]
        b: Point3[ScalarType]
        c: Point3[ScalarType]

    parsed_instance = tyro.cli(
        Triangle[float],
        args=[
            "--a.x",
            "1.0",
            "--a.y",
            "1.2",
            "--a.z",
            "1.3",
            "--a.frame",
            "WORLD",
            "--b.x",
            "1.0",
            "--b.y",
            "1.2",
            "--b.z",
            "1.3",
            "--b.frame",
            "WORLD",
            "--c.x",
            "1.0",
            "--c.y",
            "1.2",
            "--c.z",
            "1.3",
            "--c.frame",
            "WORLD",
        ],
    )
    assert parsed_instance == Triangle(
        Point3(1.0, 1.2, 1.3, CoordinateFrame.WORLD),
        Point3(1.0, 1.2, 1.3, CoordinateFrame.WORLD),
        Point3(1.0, 1.2, 1.3, CoordinateFrame.WORLD),
    )
    _check_serialization_identity(Triangle[float], parsed_instance)


def test_multilevel_generic_no_helptext() -> None:
    @dataclasses.dataclass
    class LineSegment(Generic[ScalarType]):
        a: Point3[ScalarType]
        b: Point3[ScalarType]

    f = io.StringIO()
    with pytest.raises(SystemExit):
        with contextlib.redirect_stdout(f):
            tyro.cli(LineSegment[int], args=["--help"])
    helptext = f.getvalue()

    # Check that we don't accidentally grab docstrings from the generic alias!
    assert "The central part of internal API" not in helptext


def test_generic_nested_dataclass() -> None:
    @dataclasses.dataclass
    class Child:
        a: int
        b: int

    T = TypeVar("T")

    @dataclasses.dataclass
    class DataclassGeneric(Generic[T]):
        child: T

    parsed_instance = tyro.cli(
        DataclassGeneric[Child], args=["--child.a", "5", "--child.b", "7"]
    )
    assert parsed_instance == DataclassGeneric(Child(5, 7))

    # Local generics will break.
    with pytest.raises(yaml.constructor.ConstructorError):
        _check_serialization_identity(DataclassGeneric[Child], parsed_instance)


def test_generic_nested_dataclass_helptext() -> None:
    @dataclasses.dataclass
    class Child:
        a: int
        b: int

    T = TypeVar("T")

    @dataclasses.dataclass
    class DataclassGeneric(Generic[T]):
        child: T

    f = io.StringIO()
    with pytest.raises(SystemExit):
        with contextlib.redirect_stdout(f):
            tyro.cli(DataclassGeneric[Child], args=["--help"])
    helptext = f.getvalue()

    # Check that we don't accidentally grab docstrings from the generic alias!
    assert "The central part of internal API" not in helptext


def test_generic_subparsers() -> None:
    @dataclasses.dataclass
    class CommandOne:
        a: int

    @dataclasses.dataclass
    class CommandTwo:
        b: int

    T1 = TypeVar("T1")
    T2 = TypeVar("T2")

    @dataclasses.dataclass
    class Subparser(Generic[T1, T2]):
        command: Union[T1, T2]

    parsed_instance = tyro.cli(
        Subparser[CommandOne, CommandTwo],
        args="command:command-one --command.a 5".split(" "),
    )
    # Not supported in mypy.
    assert parsed_instance == Subparser(CommandOne(5))  # type: ignore
    # Local generics will break.
    with pytest.raises(yaml.constructor.ConstructorError):
        _check_serialization_identity(
            Subparser[CommandOne, CommandTwo], parsed_instance
        )

    parsed_instance = tyro.cli(
        Subparser[CommandOne, CommandTwo],
        args="command:command-two --command.b 7".split(" "),
    )
    # Not supported in mypy.
    assert parsed_instance == Subparser(CommandTwo(7))  # type: ignore
    # Local generics will break.
    with pytest.raises(yaml.constructor.ConstructorError):
        _check_serialization_identity(
            Subparser[CommandOne, CommandTwo], parsed_instance
        )


def test_generic_subparsers_in_container() -> None:
    T = TypeVar("T")

    @dataclasses.dataclass
    class Command(Generic[T]):
        a: List[T]

    T1 = TypeVar("T1")
    T2 = TypeVar("T2")

    @dataclasses.dataclass
    class Subparser(Generic[T1, T2]):
        command: Union[T1, T2]

    parsed_instance = tyro.cli(
        Subparser[Command[int], Command[float]],
        args="command:command-int --command.a 5 3".split(" "),
    )
    # Not supported in mypy.
    assert parsed_instance == Subparser(Command([5, 3])) and isinstance(  # type: ignore
        parsed_instance.command.a[0], int
    )
    # Local generics will break.
    with pytest.raises(yaml.constructor.ConstructorError):
        _check_serialization_identity(
            Subparser[Command[int], Command[float]], parsed_instance
        )

    parsed_instance = tyro.cli(
        Subparser[Command[int], Command[float]],
        args="command:command-float --command.a 7 2".split(" "),
    )
    # Not supported in mypy.
    assert parsed_instance == Subparser(Command([7.0, 2.0])) and isinstance(  # type: ignore
        parsed_instance.command.a[0], float
    )
    # Local generics will break.
    with pytest.raises(yaml.constructor.ConstructorError):
        _check_serialization_identity(
            Subparser[Command[int], Command[float]], parsed_instance
        )


def test_serialize_missing() -> None:
    @dataclasses.dataclass
    class TupleGenericVariableMissing(Generic[ScalarType]):
        xyz: Tuple[ScalarType, ...]

    x = TupleGenericVariableMissing[int](xyz=(tyro.MISSING, tyro.MISSING))
    assert tyro.extras.to_yaml(x).count("!missing") == 2
    _check_serialization_identity(TupleGenericVariableMissing[int], x)
    assert (
        tyro.extras.from_yaml(
            TupleGenericVariableMissing[int], tyro.extras.to_yaml(x)
        ).xyz[0]
        is tyro.MISSING
    )


def test_generic_inherited_type_narrowing() -> None:
    T = TypeVar("T")

    @dataclasses.dataclass
    class ActualParentClass(Generic[T]):
        x: T  # Documentation 1

        # Documentation 2
        y: T

        z: T = 3  # type: ignore
        """Documentation 3"""

    @dataclasses.dataclass
    class ChildClass(ActualParentClass[int]):
        a: int = 7

    def main(x: ActualParentClass[int] = ChildClass(5, 5)) -> ActualParentClass:
        return x

    assert tyro.cli(main, args="--x.x 3".split(" ")) == ChildClass(3, 5, 3)


def test_pculbertson() -> None:
    # https://github.com/brentyi/tyro/issues/7
    from typing import Union

    @dataclasses.dataclass(frozen=True)
    class TypeA:
        data: int

    @dataclasses.dataclass
    class TypeB:
        data: int

    @dataclasses.dataclass
    class Wrapper:
        subclass: Union[TypeA, TypeB] = TypeA(1)

    wrapper1 = Wrapper()  # Create Wrapper object.
    assert wrapper1 == tyro.extras.from_yaml(Wrapper, tyro.extras.to_yaml(wrapper1))


def test_annotated() -> None:
    # https://github.com/brentyi/tyro/issues/7

    @dataclasses.dataclass(frozen=True)
    class TypeA:
        data: int

    @dataclasses.dataclass
    class Wrapper:
        subclass: Annotated[TypeA, int] = TypeA(1)

    wrapper1 = Wrapper()  # Create Wrapper object.
    assert wrapper1 == tyro.extras.from_yaml(Wrapper, tyro.extras.to_yaml(wrapper1))


@dataclasses.dataclass
class TypeA:
    data: int


@dataclasses.dataclass
class TypeASubclass(TypeA):
    pass


@dataclasses.dataclass
class Wrapper:
    subclass: TypeA


def test_superclass() -> None:
    # https://github.com/brentyi/tyro/issues/7

    wrapper1 = Wrapper(TypeASubclass(3))  # Create Wrapper object.
    assert wrapper1 == tyro.extras.from_yaml(Wrapper, tyro.extras.to_yaml(wrapper1))



"""Tests for unsupported union types.

Unions like `int | str` or `SomeDataclassA | SomeDataclassB` are OK (note that the latter
will produce a pair of subcommands); when we write things like `int | SomeDataclassA`
handling gets more complicated but should still be supported!
"""

import dataclasses
from typing import Any, Dict, List, Tuple, Union

import pytest

import tyro


def test_subparser_strip_non_nested() -> None:
    @dataclasses.dataclass
    class DefaultHTTPServer:
        y: int

    @dataclasses.dataclass
    class DefaultSMTPServer:
        z: int

    @dataclasses.dataclass
    class DefaultSubparser:
        x: int
        # We add [int, str] to the annotation here... this should be ignored.
        bc: Union[int, str, DefaultHTTPServer, DefaultSMTPServer] = dataclasses.field(
            default_factory=lambda: DefaultHTTPServer(5)
        )

    assert (
        tyro.cli(
            DefaultSubparser, args=["--x", "1", "bc:default-http-server", "--bc.y", "5"]
        )
        == tyro.cli(DefaultSubparser, args=["--x", "1"])
        == DefaultSubparser(x=1, bc=DefaultHTTPServer(y=5))
    )
    assert tyro.cli(
        DefaultSubparser, args=["--x", "1", "bc:default-smtp-server", "--bc.z", "3"]
    ) == DefaultSubparser(x=1, bc=DefaultSMTPServer(z=3))
    assert (
        tyro.cli(
            DefaultSubparser, args=["--x", "1", "bc:default-http-server", "--bc.y", "8"]
        )
        == tyro.cli(
            DefaultSubparser,
            args=[],
            default=DefaultSubparser(x=1, bc=DefaultHTTPServer(y=8)),
        )
        == DefaultSubparser(x=1, bc=DefaultHTTPServer(y=8))
    )

    with pytest.raises(SystemExit):
        tyro.cli(DefaultSubparser, args=["--x", "1", "b", "--bc.z", "3"])
    with pytest.raises(SystemExit):
        tyro.cli(DefaultSubparser, args=["--x", "1", "c", "--bc.y", "3"])


def test_subparser_strip_nested() -> None:
    @dataclasses.dataclass
    class DefaultHTTPServer:
        y: int

    @dataclasses.dataclass
    class DefaultSMTPServer:
        z: int

    @dataclasses.dataclass
    class DefaultSubparser:
        x: int
        # We add [int, str] to the annotation here... this should be ignored.
        bc: Union[int, str, DefaultHTTPServer, DefaultSMTPServer] = 5

    assert (
        tyro.cli(DefaultSubparser, args=["--x", "1", "bc:int", "5"])
        == tyro.cli(DefaultSubparser, args=["--x", "1"])
        == DefaultSubparser(x=1, bc=5)
    )
    assert tyro.cli(
        DefaultSubparser, args=["--x", "1", "bc:str", "five"]
    ) == DefaultSubparser(x=1, bc="five")


def test_with_fancy_types() -> None:
    @dataclasses.dataclass
    class Args:
        y: int

    def main(x: Union[Tuple[int, ...], List[str], Args, Dict[str, int]]) -> Any:
        return x

    assert tyro.cli(main, args="x:tuple-int-ellipsis 1 2 3".split(" ")) == (1, 2, 3)
    assert tyro.cli(main, args="x:list-str 1 2 3".split(" ")) == ["1", "2", "3"]
    assert tyro.cli(main, args="x:args --x.y 5".split(" ")) == Args(5)
    assert tyro.cli(main, args="x:dict-str-int 1 2 3 4".split(" ")) == {"1": 2, "3": 4}



from __future__ import annotations

import pytest
from typing_extensions import Self

import tyro


class SomeClass:
    def __init__(self, a: int, b: int) -> None:
        self.a = a
        self.b = b

    def method1(self, x: Self) -> None:
        self.effect = x

    @classmethod
    def method2(cls, x: Self) -> SomeClass:
        return x

    # Self is not valid in static methods.
    # https://peps.python.org/pep-0673/#valid-locations-for-self
    #
    # @staticmethod
    # def method3(x: Self) -> TestClass:
    #     return x


class SomeSubclass(SomeClass): ...


def test_method() -> None:
    x = SomeClass(0, 0)
    with pytest.raises(SystemExit):
        tyro.cli(x.method1, args=[])

    assert tyro.cli(x.method1, args="--x.a 3 --x.b 3".split(" ")) is None
    assert x.effect.a == 3 and x.effect.b == 3
    assert isinstance(x, SomeClass)


def test_classmethod() -> None:
    x = SomeClass(0, 0)
    with pytest.raises(SystemExit):
        tyro.cli(x.method2, args=[])
    with pytest.raises(SystemExit):
        tyro.cli(SomeClass.method2, args=[])

    y = tyro.cli(x.method2, args="--x.a 3 --x.b 3".split(" "))
    assert y.a == 3
    assert y.b == 3
    assert isinstance(y, SomeClass)

    y = tyro.cli(SomeClass.method2, args="--x.a 3 --x.b 3".split(" "))
    assert y.a == 3
    assert y.b == 3
    assert isinstance(y, SomeClass)


def test_subclass_method() -> None:
    x = SomeSubclass(0, 0)
    with pytest.raises(SystemExit):
        tyro.cli(x.method1, args=[])

    assert tyro.cli(x.method1, args="--x.a 3 --x.b 3".split(" ")) is None
    assert x.effect.a == 3 and x.effect.b == 3
    assert isinstance(x, SomeSubclass)

    y = tyro.cli(x.method2, args="--x.a 3 --x.b 3".split(" "))
    assert y.a == 3
    assert y.b == 3
    assert isinstance(y, SomeClass)


def test_subclass_classmethod() -> None:
    x = SomeSubclass(0, 0)
    with pytest.raises(SystemExit):
        tyro.cli(x.method2, args=[])
    with pytest.raises(SystemExit):
        tyro.cli(SomeSubclass.method2, args=[])

    y = tyro.cli(x.method2, args="--x.a 3 --x.b 3".split(" "))
    assert y.a == 3
    assert y.b == 3
    assert isinstance(y, SomeClass)

    y = tyro.cli(SomeSubclass.method2, args="--x.a 3 --x.b 3".split(" "))
    assert y.a == 3
    assert y.b == 3
    assert isinstance(y, SomeClass)



# mypy: ignore-errors
#
# PEP 695 isn't yet supported in mypy. (April 4, 2024)
"""Type Aliases (3.12+)

In Python 3.12, the :code:`type` statement is introduced to create type aliases.

Usage:

    python ./11_type_aliases_py312.py --help
"""

import dataclasses

import tyro

# Lazily-evaluated type alias.
type Field1Type = Inner


@dataclasses.dataclass
class Inner:
    a: int
    b: str


@dataclasses.dataclass
class Args:
    """Description.
    This should show up in the helptext!"""

    field1: Field1Type
    """A field."""

    field2: int = 3
    """A numeric field, with a default value."""


if __name__ == "__main__":
    args = tyro.cli(Args)
    print(args)



"""Abstractions for pulling out 'field' definitions, which specify inputs, types, and # type: ignore
defaults, from general callables."""

from __future__ import annotations

import contextlib
import dataclasses
import functools
import inspect
import numbers
import warnings
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

import docstring_parser
from typing_extensions import Annotated, get_args, get_origin

from . import _docstrings, _resolver, _strings, _unsafe_cache
from ._singleton import (
    DEFAULT_SENTINEL_SINGLETONS,
    MISSING_AND_MISSING_NONPROP,
    MISSING_NONPROP,
)
from ._typing import TypeForm
from .conf import _confstruct, _markers
from .constructors._primitive_spec import (
    PrimitiveTypeInfo,
    UnsupportedTypeAnnotationError,
)
from .constructors._registry import ConstructorRegistry
from .constructors._struct_spec import (
    StructFieldSpec,
    StructTypeInfo,
    UnsupportedStructTypeMessage,
)

global_context_markers: List[Tuple[_markers.Marker, ...]] = []


@dataclasses.dataclass
class FieldDefinition:
    intern_name: str
    extern_name: str
    type: TypeForm[Any] | Callable
    """Full type, including runtime annotations."""
    type_stripped: TypeForm[Any] | Callable
    default: Any
    # We need to record whether defaults are from default instances to
    # determine if they should override the default in
    # tyro.conf.subcommand(default=...).
    is_default_from_default_instance: bool
    helptext: Optional[str]
    markers: Set[Any]
    custom_constructor: bool

    argconf: _confstruct._ArgConfig

    # Override the name in our kwargs. Useful whenever the user-facing argument name
    # doesn't match the keyword expected by our callable.
    call_argname: Any

    def __post_init__(self):
        if (
            _markers.Fixed in self.markers or _markers.Suppress in self.markers
        ) and self.default in MISSING_AND_MISSING_NONPROP:
            raise UnsupportedTypeAnnotationError(
                f"Field {self.intern_name} is missing a default value!"
            )

    @staticmethod
    @contextlib.contextmanager
    def marker_context(markers: Tuple[_markers.Marker, ...]):
        """Context for setting markers on fields. All fields created within the
        context will have the specified markers."""
        global_context_markers.append(markers)
        yield
        global_context_markers.pop()

    @staticmethod
    def from_field_spec(field_spec: StructFieldSpec) -> FieldDefinition:
        return FieldDefinition.make(
            name=field_spec.name,
            typ=field_spec.type,
            default=field_spec.default,
            is_default_from_default_instance=field_spec.is_default_overridden,
            helptext=field_spec.helptext,
            call_argname_override=field_spec._call_argname,
        )

    @staticmethod
    def make(
        name: str,
        typ: Union[TypeForm[Any], Callable],
        default: Any,
        is_default_from_default_instance: bool,
        helptext: Optional[str],
        call_argname_override: Optional[Any] = None,
    ):
        # Resolve generics.
        typ = _resolver.TypeParamResolver.concretize_type_params(typ)

        # Narrow types.
        if typ is Any and default not in MISSING_AND_MISSING_NONPROP:
            typ = type(default)
        else:
            # TypeVar constraints are already applied in
            # TypeParamResolver.concretize_type_params(), but that won't be
            # called for functions.
            typ = _resolver.type_from_typevar_constraints(typ)
            typ = _resolver.narrow_collection_types(typ, default)
            typ = _resolver.narrow_union_type(typ, default)

        # Try to extract argconf overrides from type.
        _, argconfs = _resolver.unwrap_annotated(typ, _confstruct._ArgConfig)
        argconf = _confstruct._ArgConfig(
            None,
            None,
            help=None,
            help_behavior_hint=None,
            aliases=None,
            prefix_name=True,
            constructor_factory=None,
        )
        for overwrite_argconf in argconfs:
            # Apply any annotated argument configuration values.
            argconf = dataclasses.replace(
                argconf,
                **{
                    field.name: getattr(overwrite_argconf, field.name)
                    for field in dataclasses.fields(overwrite_argconf)
                    if getattr(overwrite_argconf, field.name) is not None
                },
            )
            if argconf.help is not None:
                helptext = argconf.help

        type_stripped, markers = _resolver.unwrap_annotated(typ, _markers._Marker)

        # Include markers set via context manager.
        for context_markers in global_context_markers:
            markers += context_markers

        out = FieldDefinition(
            intern_name=name,
            extern_name=name if argconf.name is None else argconf.name,
            type=typ,
            type_stripped=type_stripped,
            default=default,
            is_default_from_default_instance=is_default_from_default_instance,
            helptext=helptext,
            markers=set(markers),
            custom_constructor=argconf.constructor_factory is not None,
            argconf=argconf,
            call_argname=(
                call_argname_override if call_argname_override is not None else name
            ),
        )

        if argconf.constructor_factory is not None:
            out = out.with_new_type_stripped(argconf.constructor_factory())

        # Check that the default value matches the final resolved type.
        # There's some similar Union-specific logic for this in narrow_union_type(). We
        # may be able to consolidate this.
        if (
            # Be relatively conservative: isinstance() can be checked on non-type
            # types (like unions in Python >=3.10), but we'll only consider single types
            # for now.
            type(out.type_stripped) is type
            and not isinstance(default, out.type_stripped)  # type: ignore
            # If a custom constructor is set, static_type may not be
            # matched to the annotated type.
            and argconf.constructor_factory is None
            and default not in DEFAULT_SENTINEL_SINGLETONS
            # The numeric tower in Python is wacky. This logic is non-critical, so
            # we'll just skip it (+the complexity) for numbers.
            and not isinstance(default, numbers.Number)
        ):
            # If the default value doesn't match the resolved type, we expand the
            # type. This is inspired by https://github.com/brentyi/tyro/issues/88.
            warnings.warn(
                f"The field {name} is annotated with type {typ}, "
                f"but the default value {default} has type {type(default)}. "
                f"We'll try to handle this gracefully, but it may cause unexpected behavior."
            )
            out = out.with_new_type_stripped(Union[out.type_stripped, type(default)])  # type: ignore

        return out

    def with_new_type_stripped(
        self, new_type_stripped: TypeForm[Any] | Callable
    ) -> FieldDefinition:
        if get_origin(self.type) is Annotated:
            new_type = Annotated[(new_type_stripped, *get_args(self.type)[1:])]  # type: ignore
        else:
            new_type = new_type_stripped  # type: ignore
        return dataclasses.replace(
            self,
            type=new_type,  # type: ignore
            type_stripped=new_type_stripped,
        )

    def is_positional(self) -> bool:
        """Returns True if the argument should be positional in the commandline."""
        return (
            # Explicit positionals.
            _markers.Positional in self.markers
            # Dummy dataclasses should have a single positional field.
            or self.intern_name == _strings.dummy_field_name
            or (
                # Make required arguments positional.
                _markers.PositionalRequiredArgs in self.markers
                and self.default in MISSING_AND_MISSING_NONPROP
            )
        )

    def is_positional_call(self) -> bool:
        """Returns True if the argument should be positional in underlying Python call."""
        return (
            # Explicit positionals.
            _markers._PositionalCall in self.markers
            # Dummy dataclasses should have a single positional field.
            or self.intern_name == _strings.dummy_field_name
        )


@_unsafe_cache.unsafe_cache(maxsize=1024)
def is_struct_type(typ: Union[TypeForm[Any], Callable], default_instance: Any) -> bool:
    """Determine whether a type should be treated as a 'struct type', where a single
    type can be broken down into multiple fields (eg for nested dataclasses or
    classes).

    TODO: we should come up with a better name than 'struct type', which is a little bit
    misleading."""

    list_or_error = field_list_from_type_or_callable(
        typ, default_instance, support_single_arg_types=False
    )
    return not isinstance(
        list_or_error,
        UnsupportedStructTypeMessage,
    )


def field_list_from_type_or_callable(
    f: Union[Callable, TypeForm[Any]],
    default_instance: Any,
    support_single_arg_types: bool,
) -> (
    UnsupportedStructTypeMessage
    | tuple[Callable | TypeForm[Any], list[FieldDefinition]]
):
    """Generate a list of generic 'field' objects corresponding to the inputs of some
    annotated callable.

    Returns:
        The type that `f` is resolved as.
        A list of field definitions.
    """

    f = _resolver.swap_type_using_confstruct(f)
    registry = ConstructorRegistry._get_active_registry()
    type_info = StructTypeInfo.make(f, default_instance)

    with type_info._typevar_context:
        spec = registry.get_struct_spec(type_info)

        with FieldDefinition.marker_context(type_info.markers):
            if spec is not None:
                return f, [FieldDefinition.from_field_spec(f) for f in spec.fields]

            try:
                registry.get_primitive_spec(PrimitiveTypeInfo.make(f, set()))
                is_primitive = True
            except UnsupportedTypeAnnotationError:
                is_primitive = False

            if is_primitive and support_single_arg_types:
                with FieldDefinition.marker_context(
                    (_markers.Positional, _markers._PositionalCall)
                ):
                    return (
                        f,
                        [
                            FieldDefinition.make(
                                name="value",
                                typ=f,
                                default=default_instance,
                                is_default_from_default_instance=True,
                                helptext="",
                            )
                        ],
                    )
            elif not is_primitive and callable(f):
                return _field_list_from_function(
                    type_info.type,  # This will have typing.Annotated metadata stripped.
                    default_instance,
                )

    return UnsupportedStructTypeMessage(f"{f} is not a valid struct type!")


def _field_list_from_function(
    f: Callable, default_instance: Any
) -> UnsupportedStructTypeMessage | tuple[Callable, list[FieldDefinition]]:
    """Generate field lists from non-class callables."""
    try:
        params = list(inspect.signature(f).parameters.values())
    except ValueError:
        return UnsupportedStructTypeMessage(f"Could not get signature for {f}!")

    # Unwrap functools.wraps and functools.partial.
    done = False
    while not done:
        done = True
        if hasattr(f, "__wrapped__"):
            f = f.__wrapped__  # type: ignore
            done = False
        if isinstance(f, functools.partial):
            f = f.func
            done = False

    # Check for abstract classes.
    if inspect.isabstract(f):
        return UnsupportedStructTypeMessage("Abstract classes cannot be instantiated!")

    # `f` that is called (output) may be different from what we want to
    # inspect.
    f_out = f
    if inspect.isclass(f):
        if hasattr(f, "__init__") and f.__init__ is not object.__init__:
            f = f.__init__  # type: ignore
        elif hasattr(f, "__new__") and f.__new__ is not object.__new__:
            f = f.__new__

    # Get type annotations, docstrings.
    docstring = inspect.getdoc(f)
    docstring_from_arg_name = {}
    if docstring is not None:
        for param_doc in docstring_parser.parse(docstring).params:
            docstring_from_arg_name[param_doc.arg_name] = param_doc.description
    del docstring

    # This will throw a type error for torch.device, typing.Dict, etc.
    try:
        hints = _resolver.get_type_hints_with_backported_syntax(f, include_extras=True)
    except TypeError:
        return UnsupportedStructTypeMessage(f"Could not get hints for {f}!")

    field_list = []
    for param in params:
        # Get default value.
        default = param.default

        # Get helptext from docstring.
        helptext = docstring_from_arg_name.get(param.name)

        # TODO: re-add.
        if helptext is None and inspect.isclass(f_out):
            helptext = _docstrings.get_field_docstring(f_out, param.name)

        if param.name not in hints:
            out = UnsupportedStructTypeMessage(
                f"Expected fully type-annotated callable, but {f} with arguments"
                f" {tuple(map(lambda p: p.name, params))} has no annotation for"
                f" '{param.name}'."
            )
            if param.kind is param.KEYWORD_ONLY:
                # If keyword only: this can't possibly be an instantiator function
                # either, so we escalate to an error.
                raise UnsupportedTypeAnnotationError(out.message)
            return out

        # Set markers for positional + variadic arguments.
        markers: Tuple[Any, ...] = ()
        typ: Any = hints[param.name]
        if param.kind is inspect.Parameter.POSITIONAL_ONLY:
            markers = (_markers.Positional, _markers._PositionalCall)
        elif param.kind is inspect.Parameter.VAR_POSITIONAL:
            # Handle *args signatures.
            #
            # This will create a `--args T [T ...]` CLI argument.
            markers = (_markers._UnpackArgsCall,)
            typ = Tuple.__getitem__((typ, ...))  # type: ignore
            default = ()
        elif param.kind is inspect.Parameter.VAR_KEYWORD:
            # Handle *kwargs signatures.
            #
            # This will create a `--kwargs STR T [STR T ...]` CLI argument.
            #
            # It would be straightforward to make both this and *args truly
            # positional, omitting the --args/--kwargs prefix, but we are
            # choosing not to because it would make *args and **kwargs
            # difficult to use in conjunction.
            markers = (_markers._UnpackKwargsCall,)
            typ = Dict.__getitem__((str, typ))  # type: ignore
            default = {}

        with FieldDefinition.marker_context(markers):
            field_list.append(
                FieldDefinition.make(
                    name=param.name,
                    # param.annotation doesn't resolve forward references.
                    typ=typ
                    if default_instance in MISSING_AND_MISSING_NONPROP
                    else Annotated[(typ, _markers._OPTIONAL_GROUP)],  # type: ignore
                    default=default if default is not param.empty else MISSING_NONPROP,
                    is_default_from_default_instance=False,
                    helptext=helptext,
                )
            )

    return f_out, field_list



from typing import Generic, TypeVar

from pydantic import BaseModel

import tyro

T = TypeVar("T")


def test_pydantic_generic() -> None:
    class ManyTypesA(BaseModel, Generic[T]):
        i: T
        s: str = "hello"
        f: float = 3.0

    assert tyro.cli(ManyTypesA[int], args=["--i", "5"]) == ManyTypesA(
        i=5, s="hello", f=3.0
    )



import contextlib
import copy
import dataclasses
import io
import pathlib
from typing import (
    Any,
    Dict,
    Literal,
    Mapping,
    NamedTuple,
    NotRequired,
    Required,
    Tuple,
    TypedDict,
    cast,
)

import pytest

import tyro
import tyro._strings


def test_basic_dict() -> None:
    def main(params: Dict[str, int]) -> Dict[str, int]:
        return params

    assert tyro.cli(main, args="--params hey 5 hello 2".split(" ")) == {
        "hey": 5,
        "hello": 2,
    }
    assert tyro.cli(main, args="--params hey 5 hello 2".split(" ")) == {
        "hey": 5,
        "hello": 2,
    }
    assert tyro.cli(main, args="--params".split(" ")) == {}
    with pytest.raises(SystemExit):
        tyro.cli(main, args="--params hey 5 hello hey".split(" "))
    with pytest.raises(SystemExit):
        tyro.cli(main, args="--params hey 5 hello".split(" "))


def test_dict_with_default() -> None:
    def main(params: Mapping[Literal[1, 3, 5, 7], bool] = {5: False, 1: True}) -> Any:
        return params

    assert tyro.cli(main, args=[]) == {5: False, 1: True}
    assert tyro.cli(main, args="--params.5 --params.no-1".split(" ")) == {
        5: True,
        1: False,
    }
    with pytest.raises(SystemExit):
        tyro.cli(main, args="--params".split(" "))


def test_tuple_in_dict() -> None:
    def main(x: Dict[Tuple[int, int] | Tuple[str, str], Tuple[int, int]]) -> dict:
        return x

    assert tyro.cli(main, args="--x 1 1 2 2 3 3 4 4".split(" ")) == {
        (1, 1): (2, 2),
        (3, 3): (4, 4),
    }


def test_basic_typeddict() -> None:
    class ManyTypesTypedDict(TypedDict):
        i: int
        s: str

    assert tyro.cli(
        ManyTypesTypedDict,
        args="--i 5 --s 5".split(" "),
    ) == dict(i=5, s="5")

    with pytest.raises(SystemExit):
        tyro.cli(ManyTypesTypedDict, args="--i 5".split(" "))

    with pytest.raises(SystemExit):
        tyro.cli(ManyTypesTypedDict, args="--s 5".split(" "))


def test_positional_in_typeddict() -> None:
    class ManyTypesTypedDict(TypedDict):
        i: tyro.conf.Positional[int]
        s: str

    assert tyro.cli(
        ManyTypesTypedDict,
        args="5 --s 5".split(" "),
    ) == dict(i=5, s="5")

    with pytest.raises(SystemExit):
        tyro.cli(ManyTypesTypedDict, args="5".split(" "))

    with pytest.raises(SystemExit):
        tyro.cli(ManyTypesTypedDict, args="--s 5".split(" "))


def test_total_false_typeddict() -> None:
    class ManyTypesTypedDict(TypedDict, total=False):
        i: int
        s: str

    assert tyro.cli(
        ManyTypesTypedDict,
        args="--i 5 --s 5".split(" "),
    ) == dict(i=5, s="5")

    assert tyro.cli(ManyTypesTypedDict, args=[]) == dict()
    assert tyro.cli(ManyTypesTypedDict, args="--i 5".split(" ")) == dict(i=5)
    assert tyro.cli(ManyTypesTypedDict, args="--s 5".split(" ")) == dict(s="5")


def test_total_false_required_typeddict() -> None:
    class ManyTypesTypedDict(TypedDict, total=False):
        i: int
        s: Required[str]

    assert tyro.cli(
        ManyTypesTypedDict,
        args="--i 5 --s 5".split(" "),
    ) == dict(i=5, s="5")

    with pytest.raises(SystemExit):
        # `s` is Required[].
        assert tyro.cli(ManyTypesTypedDict, args="--i 5".split(" ")) == dict(i=5)
    assert tyro.cli(
        ManyTypesTypedDict, args="--i 5".split(" "), default={"s": "5"}
    ) == dict(i=5, s="5")
    assert tyro.cli(ManyTypesTypedDict, args="--s 5".split(" ")) == dict(s="5")


def test_total_true_not_required_typeddict() -> None:
    class ManyTypesTypedDict(TypedDict, total=True):
        i: NotRequired[int]
        s: str

    assert tyro.cli(
        ManyTypesTypedDict,
        args="--i 5 --s 5".split(" "),
    ) == dict(i=5, s="5")

    with pytest.raises(SystemExit):
        # `s` is Required[].
        assert tyro.cli(ManyTypesTypedDict, args="--i 5".split(" ")) == dict(i=5)
    assert tyro.cli(
        ManyTypesTypedDict, args="--i 5".split(" "), default={"s": "5"}
    ) == dict(i=5, s="5")
    assert tyro.cli(ManyTypesTypedDict, args="--s 5".split(" ")) == dict(s="5")


def test_total_false_nested_typeddict() -> None:
    class ChildTypedDict(TypedDict, total=False):
        i: int
        s: str

    class ParentTypedDict(TypedDict, total=False):
        child: ChildTypedDict

    assert tyro.cli(
        ParentTypedDict,
        args="--child.i 5 --child.s 5".split(" "),
    ) == {"child": {"i": 5, "s": "5"}}

    # total=False is ~ignored on the parent.
    assert tyro.cli(
        ParentTypedDict,
        args=[],
    ) == {"child": {}}


def test_total_false_typeddict_with_nested() -> None:
    @dataclasses.dataclass
    class Inner:
        j: float

    class ManyTypesTypedDict(TypedDict, total=False):
        i: int
        s: Inner

    # --s.j is (unfortunately) still required.
    with pytest.raises(SystemExit):
        tyro.cli(
            ManyTypesTypedDict,
            args="".split(" "),
        )

    assert tyro.cli(
        ManyTypesTypedDict,
        args="--s.j 5".split(" "),
    ) == {"s": Inner(5.0)}


def test_total_false_typeddict_with_tuple() -> None:
    class ManyTypesTypedDict(TypedDict, total=False):
        i: int
        s: Tuple[str, str]

    assert (
        tyro.cli(
            ManyTypesTypedDict,
            args=[],
        )
        == dict()
    )

    assert tyro.cli(
        ManyTypesTypedDict,
        args="--i 5 --s 5 5".split(" "),
    ) == dict(i=5, s=("5", "5"))


def test_nested_typeddict() -> None:
    class ChildTypedDict(TypedDict):
        y: int

    class NestedTypedDict(TypedDict):
        x: int
        b: ChildTypedDict

    assert tyro.cli(NestedTypedDict, args=["--x", "1", "--b.y", "3"]) == dict(
        x=1, b=dict(y=3)
    )
    with pytest.raises(SystemExit):
        tyro.cli(NestedTypedDict, args=["--x", "1"])


def test_helptext_and_default_typeddict() -> None:
    class HelptextTypedDict(TypedDict):
        """This docstring should be printed as a description."""

        x: int  # Documentation 1

        # Documentation 2
        y: int

        z: int
        """Documentation 3"""

    f = io.StringIO()
    with pytest.raises(SystemExit):
        with contextlib.redirect_stdout(f):
            tyro.cli(HelptextTypedDict, default={"z": 3}, args=["--help"])
    helptext = tyro._strings.strip_ansi_sequences(f.getvalue())
    assert cast(str, HelptextTypedDict.__doc__) in helptext
    assert "--x INT" in helptext
    assert "--y INT" in helptext
    assert "--z INT" in helptext
    assert "Documentation 1 (required)" in helptext
    assert "Documentation 2 (required)" in helptext
    assert "Documentation 3 (default: 3)" in helptext


def test_basic_namedtuple() -> None:
    class ManyTypesNamedTuple(NamedTuple):
        i: int
        s: str
        f: float
        p: pathlib.Path

    assert tyro.cli(
        ManyTypesNamedTuple,
        args=[
            "--i",
            "5",
            "--s",
            "5",
            "--f",
            "5",
            "--p",
            "~",
        ],
    ) == ManyTypesNamedTuple(i=5, s="5", f=5.0, p=pathlib.Path("~"))


def test_nested_namedtuple() -> None:
    class ChildNamedTuple(NamedTuple):
        y: int

    class NestedNamedTuple(NamedTuple):
        x: int
        b: ChildNamedTuple

    assert tyro.cli(
        NestedNamedTuple, args=["--x", "1", "--b.y", "3"]
    ) == NestedNamedTuple(x=1, b=ChildNamedTuple(y=3))
    with pytest.raises(SystemExit):
        tyro.cli(NestedNamedTuple, args=["--x", "1"])


def test_helptext_and_default_namedtuple() -> None:
    class HelptextNamedTupleDefault(NamedTuple):
        """This docstring should be printed as a description."""

        x: int  # Documentation 1

        # Documentation 2
        y: int

        z: int = 3
        """Documentation 3"""

    f = io.StringIO()
    with pytest.raises(SystemExit):
        with contextlib.redirect_stdout(f):
            tyro.cli(HelptextNamedTupleDefault, args=["--help"])
    helptext = tyro._strings.strip_ansi_sequences(f.getvalue())
    assert cast(str, HelptextNamedTupleDefault.__doc__) in helptext
    assert "--x INT" in helptext
    assert "--y INT" in helptext
    assert "--z INT" in helptext
    assert "Documentation 1 (required)" in helptext
    assert "Documentation 2 (required)" in helptext
    assert "Documentation 3 (default: 3)" in helptext


def test_helptext_and_default_namedtuple_alternate() -> None:
    class HelptextNamedTuple(NamedTuple):
        """This docstring should be printed as a description."""

        x: int  # Documentation 1

        # Documentation 2
        y: int

        z: int
        """Documentation 3"""

    with pytest.raises(SystemExit):
        tyro.cli(
            HelptextNamedTuple,
            default=tyro.MISSING,
            args=[],
        )

    f = io.StringIO()
    with pytest.raises(SystemExit):
        with contextlib.redirect_stdout(f):
            tyro.cli(
                HelptextNamedTuple,
                default=HelptextNamedTuple(
                    x=tyro.MISSING,
                    y=tyro.MISSING,
                    z=3,
                ),
                args=["--help"],
            )
    helptext = tyro._strings.strip_ansi_sequences(f.getvalue())
    assert cast(str, HelptextNamedTuple.__doc__) in helptext
    assert "Documentation 1 (required)" in helptext
    assert "Documentation 2 (required)" in helptext
    assert "Documentation 3" in helptext
    assert "(default: 3)" in helptext


def test_nested_dict() -> None:
    loaded_config = {
        "batch_size": 32,
        "optimizer": {
            "learning_rate": 1e-4,
            "epsilon": 1e-8,
            "scheduler": {"schedule_type": "constant"},
        },
    }
    backup_config = copy.deepcopy(loaded_config)
    overrided_config = tyro.cli(
        dict,
        default=loaded_config,
        args=[
            "--batch-size",
            "16",
            "--optimizer.scheduler.schedule_type",
            "exponential",
        ],
    )

    # Overridden config should be different from loaded config.
    assert overrided_config != loaded_config
    assert overrided_config["batch_size"] == 16
    assert overrided_config["optimizer"]["scheduler"]["schedule_type"] == "exponential"

    # Original loaded config should not be mutated.
    assert loaded_config == backup_config


def test_nested_dict_use_underscores() -> None:
    loaded_config = {
        "batch_size": 32,
        "optimizer": {
            "learning_rate": 1e-4,
            "epsilon": 1e-8,
            "scheduler": {"schedule_type": "constant"},
        },
    }
    backup_config = copy.deepcopy(loaded_config)
    overrided_config = tyro.cli(
        dict,
        default=loaded_config,
        args=[
            "--batch-size",
            "16",
            "--optimizer.scheduler.schedule-type",
            "exponential",
        ],
        use_underscores=True,
    )

    # Overridden config should be different from loaded config.
    assert overrided_config != loaded_config
    assert overrided_config["batch_size"] == 16
    assert overrided_config["optimizer"]["scheduler"]["schedule_type"] == "exponential"

    # Original loaded config should not be mutated.
    assert loaded_config == backup_config


def test_nested_dict_hyphen() -> None:
    # We do a lot of underscore <=> conversion in the code; this is just to make sure it
    # doesn't break anything!
    loaded_config = {
        "batch-size": 32,
        "optimizer": {
            "learning-rate": 1e-4,
            "epsilon": 1e-8,
            "scheduler": {"schedule-type": "constant"},
        },
    }
    backup_config = copy.deepcopy(loaded_config)
    overrided_config = tyro.cli(
        dict,
        default=loaded_config,
        args=[
            "--batch-size",
            "16",
            "--optimizer.scheduler.schedule-type",
            "exponential",
        ],
    )

    # Overridden config should be different from loaded config.
    assert overrided_config != loaded_config
    assert overrided_config["batch-size"] == 16
    assert overrided_config["optimizer"]["scheduler"]["schedule-type"] == "exponential"

    # Original loaded config should not be mutated.
    assert loaded_config == backup_config


def test_nested_dict_hyphen_use_underscores() -> None:
    # We do a lot of underscore <=> conversion in the code; this is just to make sure it
    # doesn't break anything!
    loaded_config = {
        "batch-size": 32,
        "optimizer": {
            "learning-rate": 1e-4,
            "epsilon": 1e-8,
            "scheduler": {"schedule-type": "constant"},
        },
    }
    backup_config = copy.deepcopy(loaded_config)
    overrided_config = tyro.cli(
        dict,
        default=loaded_config,
        args=[
            "--batch-size",
            "16",
            "--optimizer.scheduler.schedule-type",
            "exponential",
        ],
        use_underscores=True,
    )

    # Overridden config should be different from loaded config.
    assert overrided_config != loaded_config
    assert overrided_config["batch-size"] == 16
    assert overrided_config["optimizer"]["scheduler"]["schedule-type"] == "exponential"

    # Original loaded config should not be mutated.
    assert loaded_config == backup_config

    overrided_config = tyro.cli(
        dict,
        default=loaded_config,
        args=[
            "--batch_size",
            "16",
            "--optimizer.scheduler.schedule_type",
            "exponential",
        ],
        use_underscores=True,
    )

    # Overridden config should be different from loaded config.
    assert overrided_config != loaded_config
    assert overrided_config["batch-size"] == 16
    assert overrided_config["optimizer"]["scheduler"]["schedule-type"] == "exponential"

    # Original loaded config should not be mutated.
    assert loaded_config == backup_config


def test_nested_dict_annotations() -> None:
    loaded_config = {
        "optimizer": {
            "scheduler": {"schedule-type": "constant"},
        },
    }

    overrided_config = tyro.cli(
        dict,
        default=loaded_config,
        args=[
            "--optimizer.scheduler.schedule-type",
            "exponential",
        ],
    )
    assert overrided_config["optimizer"]["scheduler"]["schedule-type"] == "exponential"
    del overrided_config

    overrided_config = tyro.cli(
        Dict[str, Dict],
        default=loaded_config,
        args=[
            "--optimizer.scheduler.schedule-type",
            "exponential",
        ],
    )
    assert overrided_config["optimizer"]["scheduler"]["schedule-type"] == "exponential"
    del overrided_config

    overrided_config = tyro.cli(
        Dict[str, Dict[str, Dict]],
        default=loaded_config,
        args=[
            "--optimizer.scheduler.schedule-type",
            "exponential",
        ],
    )
    assert overrided_config["optimizer"]["scheduler"]["schedule-type"] == "exponential"
    del overrided_config


def test_functional_typeddict():
    """Source: https://github.com/brentyi/tyro/issues/87"""
    NerfMLPHiddenLayers_0 = TypedDict(
        "NerfMLPHiddenLayers_0",
        {
            "hidden_layers.0": int,
            "hidden_layers.1": int,
            "hidden_layers.2": int,
            "hidden_layers.3": int,
            "hidden_layers.4": int,
            "hidden_layers.5": int,
            "hidden_layers.6": int,
            "hidden_layers.7": int,
        },
    )
    NerfMLPHiddenLayers_1 = TypedDict(
        "NerfMLPHiddenLayers_1",
        {
            "hidden_layers.0": NotRequired[int],
            "hidden_layers.1": NotRequired[int],
            "hidden_layers.2": NotRequired[int],
            "hidden_layers.3": NotRequired[int],
            "hidden_layers.4": NotRequired[int],
            "hidden_layers.5": NotRequired[int],
            "hidden_layers.6": NotRequired[int],
            "hidden_layers.7": NotRequired[int],
        },
    )
    NerfMLPHiddenLayers_2 = TypedDict(
        "NerfMLPHiddenLayers_2",
        {
            "hidden_layers.0": int,
            "hidden_layers.1": int,
            "hidden_layers.2": int,
            "hidden_layers.3": int,
            "hidden_layers.4": int,
            "hidden_layers.5": int,
            "hidden_layers.6": int,
            "hidden_layers.7": int,
        },
        total=False,
    )
    with pytest.raises(SystemExit):
        tyro.cli(NerfMLPHiddenLayers_0, args=["--hidden_layers.0", "3"])
    assert tyro.cli(NerfMLPHiddenLayers_1, args=["--hidden_layers.0", "3"]) == {
        "hidden_layers.0": 3
    }
    assert tyro.cli(NerfMLPHiddenLayers_2, args=["--hidden_layers.0", "3"]) == {
        "hidden_layers.0": 3
    }


def test_functional_typeddict_with_default():
    """Source: https://github.com/brentyi/tyro/issues/87"""
    NerfMLPHiddenLayers_0 = TypedDict(
        "NerfMLPHiddenLayers_0",
        {
            "hidden_layers.0": int,
            "hidden_layers.1": int,
            "hidden_layers.2": int,
            "hidden_layers.3": int,
            "hidden_layers.4": int,
            "hidden_layers.5": int,
            "hidden_layers.6": int,
            "hidden_layers.7": int,
        },
    )
    NerfMLPHiddenLayers_1 = TypedDict(
        "NerfMLPHiddenLayers_1",
        {
            "hidden_layers.0": NotRequired[int],
            "hidden_layers.1": NotRequired[int],
            "hidden_layers.2": NotRequired[int],
            "hidden_layers.3": NotRequired[int],
            "hidden_layers.4": NotRequired[int],
            "hidden_layers.5": NotRequired[int],
            "hidden_layers.6": NotRequired[int],
            "hidden_layers.7": NotRequired[int],
        },
    )
    NerfMLPHiddenLayers_2 = TypedDict(
        "NerfMLPHiddenLayers_2",
        {
            "hidden_layers.0": int,
            "hidden_layers.1": int,
            "hidden_layers.2": int,
            "hidden_layers.3": int,
            "hidden_layers.4": int,
            "hidden_layers.5": int,
            "hidden_layers.6": int,
            "hidden_layers.7": int,
        },
        total=False,
    )
    with pytest.raises(SystemExit):
        tyro.cli(NerfMLPHiddenLayers_0, args=["--hidden_layers.0", "3"], default={})
    assert tyro.cli(
        NerfMLPHiddenLayers_1, args=["--hidden_layers.0", "3"], default={}
    ) == {"hidden_layers.0": 3}
    assert tyro.cli(
        NerfMLPHiddenLayers_2, args=["--hidden_layers.0", "3"], default={}
    ) == {"hidden_layers.0": 3}



from typing import TYPE_CHECKING, Any, Callable, TypeVar

from typing_extensions import Annotated

from .. import _singleton

# Current design issue: markers are applied recursively to nested structures, but can't
# be unapplied.

# All Annotated[T, None] values are just for static checkers. The real marker
# singletons are instantiated dynamically below.
#
# An alias could ideally be made, but SpecialForm aliases aren't well supported by
# type checkers.

T = TypeVar("T")

Positional = Annotated[T, None]
"""A type ``T`` can be annotated as ``Positional[T]`` if we want to parse it as a positional
argument."""

PositionalRequiredArgs = Annotated[T, None]
"""Make all arguments without defaults positional."""

# Private marker. For when an argument is not only positional in the CLI, but also in
# the callable.
_PositionalCall = Annotated[T, None]

# Private markers for when arguments should be passed in via *args or **kwargs.
_UnpackArgsCall = Annotated[T, None]
_UnpackKwargsCall = Annotated[T, None]

# Private marker.
_OPTIONAL_GROUP = Annotated[T, None]

# TODO: the verb tenses here are inconsistent, naming could be revisited.
# Perhaps Suppress should be Suppressed? But SuppressedFixed would be weird.

Fixed = Annotated[T, None]
"""A type ``T`` can be annotated as ``Fixed[T]`` to prevent :func:`tyro.cli`
from parsing it; a default value should be set instead. Fields that can't be
parsed with defaults will also be marked as fixed automatically."""

Suppress = Annotated[T, None]
"""A type ``T`` can be annotated as ``Suppress[T]`` to prevent :func:`tyro.cli` from parsing it, and
to prevent it from showing up in helptext."""

SuppressFixed = Annotated[T, None]
"""Hide fields that are either manually or automatically marked as fixed."""

FlagConversionOff = Annotated[T, None]
"""Turn off flag conversion for booleans with default values. Instead, types annotated
with `bool` will expect an explicit True or False.

Can be used directly on boolean annotations, ``FlagConversionOff[bool]``, or recursively
applied to nested types."""

AvoidSubcommands = Annotated[T, None]
"""Avoid creating subcommands when a default is provided for unions over nested types.
This simplifies CLI interfaces, but makes them less expressive.

Can be used directly on union types, ``AvoidSubcommands[Union[...]]``, or recursively
applied to nested types."""

ConsolidateSubcommandArgs = Annotated[T, None]
"""Consolidate arguments applied to subcommands. Makes CLI less sensitive to argument
ordering, at the cost of support for optional subcommands.

By default, :mod:`tyro` will generate a traditional CLI interface where args are applied to
the directly preceding subcommand. When we have two subcommands ``s1`` and ``s2``:


.. code-block:: bash

    python x.py {--root options} s1 {--s1 options} s2 {--s2 options}

This can be frustrating because the resulting CLI is sensitive to the positioning of
options.

To consolidate subcommands, we push arguments to the end, after all subcommands:


.. code-block:: bash

    python x.py s1 s2 {--root, s1, and s2 options}


This is more robust to reordering of options, ensuring that any new options can simply
be placed at the end of the command.
"""

OmitSubcommandPrefixes = Annotated[T, None]
"""Make CLI inputs used for subcommands shorter by omitting the subcommand-specific
portion of the prefix.

If we have a structure with the field ``cmd: Union[NestedTypeA, NestedTypeB]``:

By default, ``--cmd.arg`` may be generated as a flag for each dataclass in the union.
If subcommand prefixes are omitted, we would instead have ``--arg``.

By default, ``cmd:nested-type-a`` and ``cmd:nested-type-b`` may be generated as subcommand.
If subcommand prefixes are omitted, we would instead have ``nested-type-a`` and
``nested-type-b``.
"""

OmitArgPrefixes = Annotated[T, None]
"""Make flags used for keyword arguments shorter by omitting prefixes.

If we have a structure with the field:

.. code-block:: python

    cmd: NestedType

By default, ``--cmd.arg`` may be generated as a flag. If prefixes are omitted, we would
instead simply have ``--arg``.
"""

UseAppendAction = Annotated[T, None]
"""Use "append" actions for variable-length arguments.

Given an annotation like ``x: list[int]``, this means that ``x = [0, 1, 2]`` can be set via
the CLI syntax ``--x 0 --x 1 --x 2`` instead of the default of ``--x 0 1 2``.

The resulting syntax may be more user-friendly; for :mod:`tyro`, it also enables support for
otherwise ambiguous annotations like ``list[list[int]]``.

Can be applied to all variable-length sequences (``list[T]``, ``Sequence[T]``,
``tuple[T, ...]``, etc), including dictionaries without default values.
"""

UseCounterAction = Annotated[T, None]
"""Use "counter" actions for integer arguments. Example usage: ``verbose: UseCounterAction[int]``."""

EnumChoicesFromValues = Annotated[T, None]
"""Populate choices from enum values rather than enum names.

Example:

.. code-block:: python

    class OutputFormats(enum.StrEnum):
        JSON = enum.auto()
        PRETTY = enum.auto()
        RICH = enum.auto()
        TOML = enum.auto()

    @dataclasses.dataclass
    class Args:
        display_format: Annotated[
            OutputFormats, tyro.conf.EnumChoicesFromValues
        ] = OutputFormats.PRETTY

The above will result in ``json``, ``pretty``, ``rich``, and ``toml`` (all lowercase) as choices,
since the auto values for `StrEnum` (Python 3.11+) are lowercase transformations of the
names. Without this marker, the choices would be ``JSON``, ``PRETTY``, ``RICH``, and ``TOML``.

Enum aliases are not relevant when this marker is present. The first entry matching the
chosen value will be selected.
"""


CallableType = TypeVar("CallableType", bound=Callable)

# Dynamically generate marker singletons.
# These can be used one of two ways:
# - Marker[T]
# - Annotated[T, Marker]


class _Marker(_singleton.Singleton):
    def __getitem__(self, key):
        return Annotated[(key, self)]  # type: ignore


Marker = Any


def configure(*markers: Marker) -> Callable[[CallableType], CallableType]:
    """Decorator for applying configuration options.

    Consider using the ``config=`` argument of :func:`tyro.cli()` instead,
    which takes the same config marker objects as inputs.

    Configuration markers are implemented via :py:data:`typing.Annotated` and straightforward
    to apply to types, for example:

    .. code-block:: python

        field: tyro.conf.FlagConversionOff[bool]

    This decorator makes markers applicable to general functions as well:

    .. code-block:: python

        # Recursively apply FlagConversionOff to all fields in `main()`.
        @tyro.conf.configure(tyro.conf.FlagConversionOff)
        def main(field: bool) -> None:
            ...

    Args:
        markers: Options to apply.
    """

    def _inner(callable: CallableType) -> CallableType:
        # We'll read from __tyro_markers__ in `_resolver.unwrap_annotated()`.
        callable.__tyro_markers__ = markers  # type: ignore
        return callable

    return _inner


if not TYPE_CHECKING:

    def _make_marker(description: str) -> _Marker:
        class _InnerMarker(_Marker):
            def __repr__(self):
                return description

        return _InnerMarker()

    _dynamic_marker_types = {}
    for k, v in dict(globals()).items():
        if v == Annotated[T, None]:
            _dynamic_marker_types[k] = _make_marker(k)
    globals().update(_dynamic_marker_types)
    del _dynamic_marker_types



from tyro import _unsafe_cache


def test_unsafe_cache():
    x = 0

    @_unsafe_cache.unsafe_cache(maxsize=2)
    def f(dummy: int):
        nonlocal x
        x += 1

    f(0)
    f(0)
    f(0)
    assert x == 1
    f(1)
    f(1)
    f(1)
    assert x == 2
    f(0)
    f(0)
    f(0)
    assert x == 2
    f(2)
    f(2)
    f(2)
    assert x == 3
    f(0)
    f(0)
    f(0)
    assert x == 4



from typing import Any, Callable, Dict, Optional, Sequence, TypeVar, Union, overload

from typing_extensions import Annotated

from tyro.conf._markers import Marker

from .._cli import cli
from ..conf import subcommand

T = TypeVar("T")


@overload
def subcommand_cli_from_dict(
    subcommands: Dict[str, Callable[..., T]],
    *,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    args: Optional[Sequence[str]] = None,
    use_underscores: bool = False,
    console_outputs: bool = True,
    config: Optional[Sequence[Marker]] = None,
) -> T: ...


# TODO: hack. We prefer the above signature, which Pyright understands, but as of 1.6.1
# mypy doesn't reason about the generics properly.
@overload
def subcommand_cli_from_dict(
    subcommands: Dict[str, Callable[..., Any]],
    *,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    args: Optional[Sequence[str]] = None,
    use_underscores: bool = False,
    console_outputs: bool = True,
    config: Optional[Sequence[Marker]] = None,
) -> Any: ...


def subcommand_cli_from_dict(
    subcommands: Dict[str, Callable[..., Any]],
    *,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    args: Optional[Sequence[str]] = None,
    use_underscores: bool = False,
    console_outputs: bool = True,
    config: Optional[Sequence[Marker]] = None,
) -> Any:
    """Generate a subcommand CLI from a dictionary of functions.

    For an input like:

    .. code-block:: python

        tyro.extras.subcommand_cli_from_dict(
            {
                "checkout": checkout,
                "commit": commit,
            }
        )

    This is internally accomplished by generating and calling:

    .. code-block:: python

        from typing import Annotated, Any, Union
        import tyro

        tyro.cli(
            Union[
                Annotated[
                    Any,
                    tyro.conf.subcommand(name="checkout", constructor=checkout),
                ],
                Annotated[
                    Any,
                    tyro.conf.subcommand(name="commit", constructor=commit),
                ],
            ]
        )

    Args:
        subcommands: Dictionary that maps the subcommand name to function to call.
        prog: The name of the program printed in helptext. Mirrors argument from
            :py:class:`argparse.ArgumentParser`.
        description: Description text for the parser, displayed when the --help flag is
            passed in. If not specified, `f`'s docstring is used. Mirrors argument from
            :py:class:`argparse.ArgumentParser`.
        args: If set, parse arguments from a sequence of strings instead of the
            commandline. Mirrors argument from :py:meth:`argparse.ArgumentParser.parse_args()`.
        use_underscores: If True, use underscores as a word delimeter instead of hyphens.
            This primarily impacts helptext; underscores and hyphens are treated equivalently
            when parsing happens. We default helptext to hyphens to follow the GNU style guide.
            https://www.gnu.org/software/libc/manual/html_node/Argument-Syntax.html
        console_outputs: If set to ``False``, parsing errors and help messages will be
            supressed. This can be useful for distributed settings, where :func:`tyro.cli()`
            is called from multiple workers but we only want console outputs from the
            main one.
        config: Sequence of config marker objects, from :mod:`tyro.conf`.
    """
    # We need to form a union type, which requires at least two elements.
    assert len(subcommands) >= 2, "At least two subcommands are required."
    return cli(
        Union.__getitem__(  # type: ignore
            tuple(
                [
                    Annotated[
                        # The constructor function can return any object.
                        Any,
                        # We'll instantiate this object by invoking a subcommand with
                        # name k, via a constructor.
                        subcommand(name=k, constructor=v),
                    ]
                    for k, v in subcommands.items()
                ]
            )
        ),
        prog=prog,
        description=description,
        args=args,
        use_underscores=use_underscores,
        console_outputs=console_outputs,
        config=config,
    )



import dataclasses
import pathlib
from typing import Any, Dict, List, Tuple

from tyro._fields import is_struct_type
from tyro._singleton import MISSING_NONPROP


def test_is_struct_type_simple():
    assert not is_struct_type(int, MISSING_NONPROP)
    assert not is_struct_type(bool, MISSING_NONPROP)
    assert not is_struct_type(str, MISSING_NONPROP)
    assert not is_struct_type(pathlib.Path, MISSING_NONPROP)


def test_is_struct_type_containers():
    assert not is_struct_type(List[int], MISSING_NONPROP)
    assert not is_struct_type(List[bool], MISSING_NONPROP)
    assert not is_struct_type(List[str], MISSING_NONPROP)
    assert not is_struct_type(List[pathlib.Path], MISSING_NONPROP)


@dataclasses.dataclass
class Color:
    r: int
    g: int
    b: int


def test_is_struct_type_actually_struct():
    assert is_struct_type(Color, Color(255, 0, 0))


def test_is_struct_type_actually_struct_narrowing():
    assert is_struct_type(Any, Color(255, 0, 0))
    assert is_struct_type(object, Color(255, 0, 0))
    assert not is_struct_type(int, Color(255, 0, 0))


def test_is_struct_type_actually_struct_in_container():
    assert is_struct_type(Tuple[Color, Color], MISSING_NONPROP)
    assert is_struct_type(Tuple[object, ...], (Color(255, 0, 0),))
    assert is_struct_type(Tuple[Any, ...], (Color(255, 0, 0),))
    assert is_struct_type(tuple, (Color(255, 0, 0),))
    assert not is_struct_type(tuple, (1, 2, 3))
    assert is_struct_type(tuple, (1, Color(255, 0, 0), 3))
    assert is_struct_type(List[Any], [Color(255, 0, 0)])


def test_struct_dict():
    assert is_struct_type(Dict[str, int], {"x": 5})
    assert is_struct_type(dict, {"x": 5})
    assert is_struct_type(Any, {"x": 5})



from __future__ import annotations

import collections.abc
import dataclasses
import enum
import sys
import warnings
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Sequence, Union

from typing_extensions import (
    Annotated,
    NotRequired,
    Required,
    cast,
    get_args,
    get_origin,
    is_typeddict,
)

from .. import _docstrings, _resolver
from .._singleton import (
    EXCLUDE_FROM_CALL,
    MISSING,
    MISSING_AND_MISSING_NONPROP,
    MISSING_NONPROP,
)
from .._typing import TypeForm
from ..conf import _confstruct, _markers

if TYPE_CHECKING:
    from ._registry import ConstructorRegistry


@dataclasses.dataclass(frozen=True)
class UnsupportedStructTypeMessage:
    """Reason why a callable cannot be treated as a struct type."""

    message: str


@dataclasses.dataclass(frozen=True)
class StructFieldSpec:
    """Behavior specification for a single field in our callable."""

    name: str
    """The name of the field. This will be used as a keyword argument for the
    struct's associated `instantiate(**kwargs)` function."""
    type: TypeForm
    """The type of the field. Can be either a primitive or a nested struct type."""
    default: Any
    """The default value of the field."""
    is_default_overridden: bool = False
    """Whether the default value was overridden by the default instance. Should
    be set to False if the default value was assigned by the field itself."""
    helptext: str | None = None
    """Helpjext for the field."""
    # TODO: it's theoretically possible to override the argname with `None`.
    _call_argname: Any = None
    """Private: the name of the argument to pass to the callable. This is used
    for dictionary types."""


@dataclasses.dataclass(frozen=True)
class StructConstructorSpec:
    """Specification for a struct type, which is broken down into multiple
    fields.

    Each struct type is instantiated by calling an `instantiate(**kwargs)`
    function with keyword a set of keyword arguments.

    Unlike `PrimitiveConstructorSpec`, there is only one way to use this class.
    It must be returned by a rule in `ConstructorRegistry`.
    """

    instantiate: Callable[..., Any]
    """Function to call to instantiate the struct."""
    fields: tuple[StructFieldSpec, ...]
    """Fields used to construct the callable. Each field is used as a keyword
    argument for the `instantiate(**kwargs)` function."""


@dataclasses.dataclass(frozen=True)
class StructTypeInfo:
    """Information used to generate constructors for struct types."""

    type: TypeForm
    """The type of the (potential) struct."""
    markers: tuple[Any, ...]
    """Markers from :mod:`tyro.conf` that are associated with this field."""
    default: Any
    """The default value of the struct, or a member of
    :data:`tyro.constructors.MISSING_SINGLETONS` if not present. In a function
    signature, this is `X` in `def main(x=X): ...`. This can be useful for
    populating the default values of the struct."""
    _typevar_context: _resolver.TypeParamAssignmentContext

    @staticmethod
    def make(f: TypeForm | Callable, default: Any) -> StructTypeInfo:
        _, parent_markers = _resolver.unwrap_annotated(f, _markers._Marker)
        f = _resolver.swap_type_using_confstruct(f)
        f, found_subcommand_configs = _resolver.unwrap_annotated(
            f, _confstruct._SubcommandConfig
        )
        if len(found_subcommand_configs) > 0:
            default = found_subcommand_configs[0].default

        # Handle generics.
        typevar_context = _resolver.TypeParamResolver.get_assignment_context(f)
        f = typevar_context.origin_type
        f = _resolver.narrow_subtypes(f, default)
        f = _resolver.narrow_collection_types(f, default)

        return StructTypeInfo(
            cast(TypeForm, f), parent_markers, default, typevar_context
        )


def apply_default_struct_rules(registry: ConstructorRegistry) -> None:
    from .._fields import is_struct_type

    @registry.struct_rule
    def dataclass_rule(info: StructTypeInfo) -> StructConstructorSpec | None:
        if not dataclasses.is_dataclass(info.type):
            return None

        is_flax_module = False
        try:
            # Check if dataclass is a flax module. This is only possible if flax is already
            # loaded.
            #
            # We generally want to avoid importing flax, since it requires a lot of heavy
            # imports.
            if "flax.linen" in sys.modules.keys():
                import flax.linen

                if issubclass(info.type, flax.linen.Module):
                    is_flax_module = True
        except ImportError:
            pass

        # Handle dataclasses.
        field_list = []
        for dc_field in filter(
            lambda field: field.init, _resolver.resolved_fields(info.type)
        ):
            # For flax modules, we ignore the built-in "name" and "parent" fields.
            if is_flax_module and dc_field.name in ("name", "parent"):
                continue

            default, is_default_from_default_instance = _get_dataclass_field_default(
                dc_field, info.default
            )

            # Try to get helptext from field metadata. This is also intended to be
            # compatible with HuggingFace-style config objects.
            helptext = dc_field.metadata.get("help", None)
            assert isinstance(helptext, (str, type(None)))

            # Try to get helptext from docstrings. This can't be generated
            # dynamically.
            if helptext is None:
                helptext = _docstrings.get_field_docstring(info.type, dc_field.name)

            assert not isinstance(dc_field.type, str)
            field_list.append(
                StructFieldSpec(
                    name=dc_field.name,
                    type=dc_field.type,
                    default=default,
                    is_default_overridden=is_default_from_default_instance,
                    helptext=helptext,
                )
            )
        return StructConstructorSpec(instantiate=info.type, fields=tuple(field_list))

    @registry.struct_rule
    def typeddict_rule(info: StructTypeInfo) -> StructConstructorSpec | None:
        # Is this a TypedDict?
        if not is_typeddict(info.type):
            return None

        cls = cast(type, info.type)

        # Handle TypedDicts.
        field_list = []
        valid_default_instance = (
            info.default not in MISSING_AND_MISSING_NONPROP
            and info.default is not EXCLUDE_FROM_CALL
        )
        assert not valid_default_instance or isinstance(info.default, dict)
        total = getattr(cls, "__total__", True)
        assert isinstance(total, bool)
        assert not valid_default_instance or isinstance(info.default, dict)
        for name, typ in _resolver.get_type_hints_with_backported_syntax(
            cls, include_extras=True
        ).items():
            typ_origin = get_origin(typ)
            is_default_from_default_instance = False
            if valid_default_instance and name in cast(dict, info.default):
                default = cast(dict, info.default)[name]
                is_default_from_default_instance = True
            elif typ_origin is Required and total is False:
                # Support total=False.
                default = MISSING
            elif total is False:
                # Support total=False.
                default = EXCLUDE_FROM_CALL
                if is_struct_type(typ, MISSING_NONPROP):
                    # total=False behavior is unideal for nested structures.
                    pass
                    # raise _instantiators.UnsupportedTypeAnnotationError(
                    #     "`total=False` not supported for nested structures."
                    # )
            elif typ_origin is NotRequired:
                # Support typing.NotRequired[].
                default = EXCLUDE_FROM_CALL
            else:
                default = MISSING

            # Nested types need to be populated / can't be excluded from the call.
            if default is EXCLUDE_FROM_CALL and is_struct_type(typ, MISSING_NONPROP):
                default = MISSING_NONPROP

            if typ_origin in (Required, NotRequired):
                args = get_args(typ)
                assert (
                    len(args) == 1
                ), "typing.Required[] and typing.NotRequired[T] require a concrete type T."
                typ = args[0]
                del args

            field_list.append(
                StructFieldSpec(
                    name=name,
                    type=typ,
                    default=default,
                    is_default_overridden=is_default_from_default_instance,
                    helptext=_docstrings.get_field_docstring(cls, name),
                )
            )
        return StructConstructorSpec(instantiate=info.type, fields=tuple(field_list))

    @registry.struct_rule
    def attrs_rule(info: StructTypeInfo) -> StructConstructorSpec | None:
        # attr will already be imported if it's used.
        if "attr" not in sys.modules.keys():  # pragma: no cover
            return None

        try:
            import attr
        except ImportError:
            # This is needed for the mock import test in
            # test_missing_optional_packages.py to pass.
            return None

        if not attr.has(info.type):
            return None

        # Resolve forward references in-place, if any exist.
        attr.resolve_types(info.type)

        # Handle attr classes.
        field_list = []
        for attr_field in attr.fields(info.type):
            # Skip fields with init=False.
            if not attr_field.init:
                continue

            # Default handling.
            name = attr_field.name
            default = attr_field.default
            is_default_from_default_instance = False
            if info.default not in MISSING_AND_MISSING_NONPROP:
                if hasattr(info.default, name):
                    default = getattr(info.default, name)
                    is_default_from_default_instance = True
                else:
                    warnings.warn(
                        f"Could not find field {name} in default instance"
                        f" {info.default}, which has"
                        f" type {type(info.default)},",
                        stacklevel=2,
                    )
            elif default is attr.NOTHING:
                default = MISSING_NONPROP
            elif isinstance(default, attr.Factory):  # type: ignore
                default = default.factory()  # type: ignore

            assert attr_field.type is not None, attr_field
            field_list.append(
                StructFieldSpec(
                    name=name,
                    type=attr_field.type,
                    default=default,
                    is_default_overridden=is_default_from_default_instance,
                    helptext=_docstrings.get_field_docstring(info.type, name),
                )
            )
        return StructConstructorSpec(instantiate=info.type, fields=tuple(field_list))

    @registry.struct_rule
    def dict_rule(info: StructTypeInfo) -> StructConstructorSpec | None:
        if is_typeddict(info.type) or (
            info.type
            not in (
                Dict,
                dict,
                collections.abc.Mapping,
            )
            and get_origin(info.type)
            not in (
                dict,
                collections.abc.Mapping,
            )
        ):
            return None

        if info.default in MISSING_AND_MISSING_NONPROP or len(info.default) == 0:
            return None

        field_list = []
        for k, v in info.default.items():
            field_list.append(
                StructFieldSpec(
                    name=str(k) if not isinstance(k, enum.Enum) else k.name,
                    type=type(v),
                    default=v,
                    is_default_overridden=True,
                    helptext=None,
                    _call_argname=k,
                )
            )
        return StructConstructorSpec(instantiate=dict, fields=tuple(field_list))

    @registry.struct_rule
    def namedtuple_rule(info: StructTypeInfo) -> StructConstructorSpec | None:
        if not (
            isinstance(info.type, type)
            and issubclass(info.type, tuple)
            and hasattr(info.type, "_fields")
        ):
            return None

        field_list = []
        field_defaults = getattr(info.type, "_field_defaults", {})

        for name, typ in _resolver.get_type_hints_with_backported_syntax(
            info.type, include_extras=True
        ).items():
            default = field_defaults.get(name, MISSING_NONPROP)
            is_default_from_default_instance = False

            if info.default not in MISSING_AND_MISSING_NONPROP and hasattr(
                info.default, name
            ):
                default = getattr(info.default, name)
                is_default_from_default_instance = True
            elif info.default is MISSING:
                default = MISSING

            field_list.append(
                StructFieldSpec(
                    name=name,
                    type=typ,
                    default=default,
                    is_default_overridden=is_default_from_default_instance,
                    helptext=_docstrings.get_field_docstring(info.type, name),
                )
            )

        return StructConstructorSpec(instantiate=info.type, fields=tuple(field_list))

    @registry.struct_rule
    def sequence_rule(info: StructTypeInfo) -> StructConstructorSpec | None:
        if get_origin(info.type) not in (
            list,
            set,
            tuple,
            Sequence,
            collections.abc.Sequence,
        ) or not isinstance(info.default, Iterable):
            return None

        contained_type = get_args(info.type)[0] if get_args(info.type) else Any

        if all(not is_struct_type(type(x), x) for x in info.default):
            return None

        field_list = []
        for i, default_i in enumerate(info.default):
            field_list.append(
                StructFieldSpec(
                    name=str(i),
                    type=cast(type, contained_type),
                    default=default_i,
                    is_default_overridden=True,
                    helptext="",
                )
            )

        return StructConstructorSpec(
            instantiate=type(info.default), fields=tuple(field_list)
        )

    @registry.struct_rule
    def tuple_rule(info: StructTypeInfo) -> StructConstructorSpec | None:
        # It's important that this tuple rule is defined *after* the general sequence rule. It should take precedence.
        if info.type is not tuple and get_origin(info.type) is not tuple:
            return None

        # Fixed-length tuples.
        children = get_args(info.type)
        if Ellipsis in children:
            return None  # We don't handle variable-length tuples here

        # Infer more specific type when tuple annotation isn't subscripted.
        if len(children) == 0:
            if info.default in MISSING_AND_MISSING_NONPROP:
                return None
            else:
                assert isinstance(info.default, tuple)
                children = tuple(type(x) for x in info.default)

        if (
            info.default in MISSING_AND_MISSING_NONPROP
            or info.default is EXCLUDE_FROM_CALL
        ):
            default_instance = (info.default,) * len(children)
        else:
            default_instance = info.default

        field_list: list[StructFieldSpec] = []
        for i, child in enumerate(children):
            default_i = default_instance[i]
            field_list.append(
                StructFieldSpec(
                    name=str(i),
                    type=child,
                    default=default_i,
                    is_default_overridden=True,
                    helptext="",
                )
            )

        contains_nested = False
        for field in field_list:
            # Inefficient, since is_struct_type will compute StructTypeInfo again.
            field_info = StructTypeInfo.make(field.type, field.default)
            if get_origin(field_info.type) is Union:
                for option in get_args(field_info.type):
                    # The second argument here is the default value, which can help with
                    # narrowing but is generall not necessary.
                    contains_nested |= is_struct_type(option, MISSING_NONPROP)
            contains_nested |= is_struct_type(field.type, field.default)
            if contains_nested:
                break

        if not contains_nested:
            return None
        return StructConstructorSpec(instantiate=tuple, fields=tuple(field_list))

    @registry.struct_rule
    def pydantic_rule(info: StructTypeInfo) -> StructConstructorSpec | None:
        # Check if pydantic is imported
        if "pydantic" not in sys.modules.keys():  # pragma: no cover
            return None

        try:
            import pydantic
        except ImportError:
            # Needed for the mock import test in
            # test_missing_optional_packages.py to pass.
            return None

        try:
            if "pydantic.v1" in sys.modules.keys():
                from pydantic import v1 as pydantic_v1
            else:  # pragma: no cover
                pydantic_v1 = None  # type: ignore
        except ImportError:
            pydantic_v1 = None  # type: ignore

        # Check if the type is a Pydantic model
        try:
            if not (
                issubclass(info.type, pydantic.BaseModel)
                or (
                    pydantic_v1 is not None
                    and issubclass(info.type, pydantic_v1.BaseModel)
                )
            ):
                return None
        except TypeError:
            # issubclass failed!
            return None

        field_list = []
        pydantic_version = int(
            getattr(pydantic, "__version__", "1.0.0").partition(".")[0]
        )

        if pydantic_version < 2 or (
            pydantic_v1 is not None and issubclass(info.type, pydantic_v1.BaseModel)
        ):
            # Pydantic 1.xx
            cls_cast = info.type
            hints = _resolver.get_type_hints_with_backported_syntax(
                info.type, include_extras=True
            )
            for pd1_field in cast(Dict[str, Any], cls_cast.__fields__).values():
                helptext = pd1_field.field_info.description
                if helptext is None:
                    helptext = _docstrings.get_field_docstring(
                        info.type, pd1_field.name
                    )

                default, is_default_from_default_instance = (
                    _get_pydantic_v1_field_default(
                        pd1_field.name, pd1_field, info.default
                    )
                )
                field_list.append(
                    StructFieldSpec(
                        name=pd1_field.name,
                        type=hints[pd1_field.name],
                        default=default,
                        is_default_overridden=is_default_from_default_instance,
                        helptext=helptext,
                    )
                )
        else:
            # Pydantic 2.xx
            for name, pd2_field in cast(Any, info.type).model_fields.items():
                helptext = pd2_field.description
                if helptext is None:
                    helptext = _docstrings.get_field_docstring(info.type, name)

                default, is_default_from_default_instance = (
                    _get_pydantic_v2_field_default(name, pd2_field, info.default)
                )
                field_list.append(
                    StructFieldSpec(
                        name=name,
                        type=(
                            Annotated[  # type: ignore
                                (pd2_field.annotation,) + tuple(pd2_field.metadata)
                            ]
                            if len(pd2_field.metadata) > 0
                            else pd2_field.annotation
                        ),
                        default=default,
                        is_default_overridden=is_default_from_default_instance,
                        helptext=helptext,
                    )
                )

        return StructConstructorSpec(instantiate=info.type, fields=tuple(field_list))


def _ensure_dataclass_instance_used_as_default_is_frozen(
    field: dataclasses.Field, default_instance: Any
) -> None:
    """Ensure that a dataclass type used directly as a default value is marked as
    frozen."""
    assert dataclasses.is_dataclass(default_instance)
    cls = type(default_instance)
    if not cls.__dataclass_params__.frozen:  # type: ignore
        warnings.warn(
            f"Mutable type {cls} is used as a default value for `{field.name}`. This is"
            " dangerous! Consider using `dataclasses.field(default_factory=...)` or"
            f" marking {cls} as frozen."
        )


def _get_dataclass_field_default(
    field: dataclasses.Field, parent_default_instance: Any
) -> tuple[Any, bool]:
    """Helper for getting the default instance for a dataclass field."""
    # If the dataclass's parent is explicitly marked MISSING, mark this field as missing
    # as well.
    if parent_default_instance is MISSING:
        return MISSING, False

    # Try grabbing default from parent instance.
    if (
        parent_default_instance not in MISSING_AND_MISSING_NONPROP
        and parent_default_instance is not None
    ):
        # Populate default from some parent, eg `default=` in `tyro.cli()`.
        if hasattr(parent_default_instance, field.name):
            return getattr(parent_default_instance, field.name), True
        else:
            warnings.warn(
                f"Could not find field {field.name} in default instance"
                f" {parent_default_instance}, which has"
                f" type {type(parent_default_instance)},",
                stacklevel=2,
            )

    # Try grabbing default from dataclass field.
    if (
        field.default not in MISSING_AND_MISSING_NONPROP
        and field.default is not dataclasses.MISSING
    ):
        default = field.default
        # dataclasses.is_dataclass() will also return true for dataclass
        # _types_, not just instances.
        if type(default) is not type and dataclasses.is_dataclass(default):
            _ensure_dataclass_instance_used_as_default_is_frozen(field, default)
        return default, False

    # Populate default from `dataclasses.field(default_factory=...)`.
    if field.default_factory is not dataclasses.MISSING and not (
        # Special case to ignore default_factory if we write:
        # `field: Dataclass = dataclasses.field(default_factory=Dataclass)`.
        #
        # In other words, treat it the same way as: `field: Dataclass`.
        #
        # The only time this matters is when we our dataclass has a `__post_init__`
        # function that mutates the dataclass. We choose here to use the default values
        # before this method is called.
        dataclasses.is_dataclass(field.type) and field.default_factory is field.type
    ):
        return field.default_factory(), False

    # Otherwise, no default.
    return MISSING_NONPROP, False


if TYPE_CHECKING:
    import pydantic as pydantic
    import pydantic.v1.fields as pydantic_v1_fields


def _get_pydantic_v1_field_default(
    name: str,
    field: pydantic_v1_fields.ModelField,
    parent_default_instance: Any,
) -> tuple[Any, bool]:
    """Helper for getting the default instance for a Pydantic field."""

    # Try grabbing default from parent instance.
    if (
        parent_default_instance not in MISSING_AND_MISSING_NONPROP
        and parent_default_instance is not None
    ):
        # Populate default from some parent, eg `default=` in `tyro.cli()`.
        if hasattr(parent_default_instance, name):
            return getattr(parent_default_instance, name), True
        else:
            warnings.warn(
                f"Could not find field {name} in default instance"
                f" {parent_default_instance}, which has"
                f" type {type(parent_default_instance)},",
                stacklevel=2,
            )

    if not field.required:
        return field.get_default(), False

    # Otherwise, no default.
    return MISSING_NONPROP, False


def _get_pydantic_v2_field_default(
    name: str,
    field: pydantic.fields.FieldInfo,
    parent_default_instance: Any,
) -> tuple[Any, bool]:
    """Helper for getting the default instance for a Pydantic field."""

    # Try grabbing default from parent instance.
    if (
        parent_default_instance not in MISSING_AND_MISSING_NONPROP
        and parent_default_instance is not None
    ):
        # Populate default from some parent, eg `default=` in `tyro.cli()`.
        if hasattr(parent_default_instance, name):
            return getattr(parent_default_instance, name), True
        else:
            warnings.warn(
                f"Could not find field {name} in default instance"
                f" {parent_default_instance}, which has"
                f" type {type(parent_default_instance)},",
                stacklevel=2,
            )

    if not field.is_required():
        return field.get_default(call_default_factory=True), False

    # Otherwise, no default.
    return MISSING_NONPROP, False



from __future__ import annotations  # Should enable support for all versions of Python.

from typing import Any, Literal

import pytest

import tyro


def test_union_basic():
    def main(x: int | str) -> int | str:
        return x

    assert tyro.cli(main, args=["--x", "5"]) == 5
    assert tyro.cli(main, args=["--x", "6"]) == 6
    assert tyro.cli(main, args=["--x", "five"]) == "five"


def test_union_with_list():
    def main(x: int | str | list[bool]) -> Any:
        return x

    assert tyro.cli(main, args=["--x", "5"]) == 5
    assert tyro.cli(main, args=["--x", "6"]) == 6
    assert tyro.cli(main, args=["--x", "five"]) == "five"
    assert tyro.cli(main, args=["--x", "True"]) == "True"
    assert tyro.cli(main, args=["--x", "True", "False"]) == [True, False]


def test_union_literal():
    def main(x: Literal[1, 2] | Literal[3, 4, 5] | str) -> int | str:
        return x

    assert tyro.cli(main, args=["--x", "5"]) == 5
    assert tyro.cli(main, args=["--x", "6"]) == "6"
    assert tyro.cli(main, args=["--x", "five"]) == "five"


def test_super_nested():
    def main(
        x: (
            None
            | list[
                tuple[
                    None | int,
                    Literal[3, 4],
                    tuple[int, int] | tuple[str, str],
                ]
            ]
        ) = None,
    ) -> Any:
        return x

    assert tyro.cli(main, args=[]) is None
    assert tyro.cli(main, args="--x None".split(" ")) is None
    assert tyro.cli(main, args="--x None 3 2 2".split(" ")) == [(None, 3, (2, 2))]
    assert tyro.cli(main, args="--x 2 3 x 2".split(" ")) == [(2, 3, ("x", "2"))]
    assert tyro.cli(main, args="--x 2 3 x 2 2 3 1 2".split(" ")) == [
        (2, 3, ("x", "2")),
        (2, 3, (1, 2)),
    ]
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--help"])



import argparse
import contextlib
import io
import os
from typing import Any, Callable, List

import pytest

import tyro
import tyro._arguments
import tyro._strings


def get_helptext_with_checks(
    f: Callable,
    args: List[str] = ["--help"],
    use_underscores: bool = False,
    default: Any = None,
) -> str:
    """Get the helptext for a given tyro with input, while running various
    checks along the way."""
    # Should be empty.
    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(
            f,
            args=args,
            use_underscores=use_underscores,
            default=default,
            console_outputs=False,
        )
    assert target.getvalue() == ""

    # Check tyro.extras.get_parser().
    parser = tyro.extras.get_parser(f, use_underscores=use_underscores)
    assert isinstance(parser, argparse.ArgumentParser)

    # Returned parser should have formatting information stripped. External tools rarely
    # support ANSI sequences.
    unformatted_helptext = parser.format_help()
    assert (
        tyro._strings.strip_ansi_sequences(unformatted_helptext) == unformatted_helptext
    ), (
        tyro._strings.strip_ansi_sequences(unformatted_helptext)
        + "\n|\n"
        + unformatted_helptext
    )
    unformatted_usage = parser.format_usage()
    assert tyro._strings.strip_ansi_sequences(unformatted_usage) == unformatted_usage

    # Basic checks for completion scripts.
    with pytest.raises(SystemExit):
        tyro.cli(
            f, default=default, args=["--tyro-write-completion", "bash", os.devnull]
        )
    for shell in ["bash", "zsh"]:
        for command in ["--tyro-print-completion", "--tyro-write-completion"]:
            target = io.StringIO()
            with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
                if command == "--tyro-write-completion":
                    tyro.cli(f, default=default, args=[command, shell, "-"])
                else:
                    # `--tyro-print-completion` is deprecated! We should use `--tyro-write-completion` instead.
                    tyro.cli(f, default=default, args=[command, shell])
            output = target.getvalue()
            assert "shtab" in output

    # Test with underscores
    for shell in ["bash", "zsh"]:
        target = io.StringIO()
        with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
            tyro.cli(
                f,
                default=default,
                args=["--tyro_write_completion", shell, "-"],
                use_underscores=True,
            )
        output = target.getvalue()
        assert "shtab" in output

    # Get the actual helptext.
    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(f, args=args, use_underscores=use_underscores, default=default)

    # Check helptext with vs without formatting. This can help catch text wrapping bugs
    # caused by ANSI sequences.
    target2 = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target2):
        tyro._arguments.USE_RICH = False
        tyro.cli(f, default=default, args=args, use_underscores=use_underscores)
        tyro._arguments.USE_RICH = True

    if target2.getvalue() != tyro._strings.strip_ansi_sequences(target.getvalue()):
        raise AssertionError(
            "Potential wrapping bug! These two strings should match:\n"
            + target2.getvalue()
            + "\n\n"
            + tyro._strings.strip_ansi_sequences(target.getvalue())
        )

    return target2.getvalue()



import enum
from typing import Iterable, TypeVar, Union

from typing_extensions import Literal

from .._typing import TypeForm

T = TypeVar("T", bound=Union[int, str, bool, enum.Enum])


def literal_type_from_choices(choices: Iterable[T]) -> TypeForm[T]:
    """Generate a :py:data:`typing.Literal` type that constrains values to a set of choices.

    Using ``Literal[...]`` directly should generally be preferred, but this function can be
    helpful when choices are generated dynamically.

    .. warning::

        The type returned by this function can be safely used as an input to
        :func:`tyro.cli()`, but for static analysis when used for annotations we
        recommend applying a `TYPE_CHECKING` guard:

        .. code-block:: python

            from typing import TYPE_CHECKING

            if TYPE_CHECKING:
                # Static type seen by language servers, type checkers, etc.
                Color = str
            else:
                # Runtime type used by tyro.
                Color = literal_type_from_choices(["red", "green", "blue"])

    Args:
        choices: Options to choose from.

    Returns:
        A type that can be passed to :func:`tyro.cli()`.
    """
    return Literal.__getitem__(tuple(choices))  # type: ignore



# mypy: ignore-errors
#
# PEP 695 isn't yet supported in mypy. (April 4, 2024)
from dataclasses import dataclass
from typing import Annotated, Any, Literal, NewType

import pytest

import tyro
from tyro.conf._markers import OmitArgPrefixes
from tyro.constructors import UnsupportedTypeAnnotationError


def test_simple_generic():
    @dataclass(frozen=True)
    class Container[T]:
        a: T
        b: T

    assert tyro.cli(Container[int], args="--a 1 --b 2".split(" ")) == Container(1, 2)


type X = int
type Y = list[int]
type Z = Inner[int]


@dataclass(frozen=True)
class Inner[T]:
    a: T
    b: T


def test_generic_with_type_statement_0():
    @dataclass(frozen=True)
    class Container[T]:
        a: T
        b: T

    assert tyro.cli(Container[X], args="--a 1 --b 2".split(" ")) == Container(1, 2)


def test_generic_with_type_statement_1():
    @dataclass(frozen=True)
    class Container[T]:
        a: tuple[X, ...]
        b: T

    assert tyro.cli(Container[Y], args="--a 1 --b 2".split(" ")) == Container((1,), [2])


def test_generic_with_type_statement_2():
    @dataclass(frozen=True)
    class Container[T]:
        a: Z

    assert tyro.cli(Container[Y], args="--a.a 1 --a.b 2".split(" ")) == Container(
        Inner(1, 2)
    )


type AnnotatedBasic = Annotated[int, tyro.conf.arg(name="basic")]


def test_annotated_alias():
    @dataclass(frozen=True)
    class Container:
        a: AnnotatedBasic

    assert tyro.cli(Container, args="--basic 1".split(" ")) == Container(1)


type TT[T] = Annotated[T, tyro.conf.arg(name="", constructor=lambda: True)]


def test_pep695_generic_alias() -> None:
    """Adapted from: https://github.com/brentyi/tyro/issues/177"""

    @dataclass(frozen=True)
    class Config:
        arg: TT[bool]

    assert tyro.cli(Config, args=[]) == Config(arg=True)


type Renamed[T] = Annotated[T, tyro.conf.arg(name="renamed")]


def test_pep695_generic_alias_rename() -> None:
    """Adapted from: https://github.com/brentyi/tyro/issues/177"""

    @dataclass(frozen=True)
    class Config:
        arg: Renamed[bool]

    assert tyro.cli(Config, args=["--renamed", "True"]) == Config(arg=True)


def test_pep695_generic_alias_rename_override() -> None:
    """Adapted from: https://github.com/brentyi/tyro/issues/177"""

    @dataclass(frozen=True)
    class Config:
        arg: Annotated[Renamed[bool], tyro.conf.arg(name="renamed2")]

    assert tyro.cli(Config, args=["--renamed2", "True"]) == Config(arg=True)


type RenamedTwice[T] = Renamed[Renamed[T]]


def test_pep695_generic_alias_rename_twice() -> None:
    """Adapted from: https://github.com/brentyi/tyro/issues/177"""

    @dataclass(frozen=True)
    class Config:
        arg: RenamedTwice[bool]

    assert tyro.cli(Config, args=["--renamed", "True"]) == Config(arg=True)


type SomeUnion = bool | int
type RenamedThreeTimes[T] = Renamed[Renamed[Renamed[T]]]
type SomeUnionRenamed = RenamedThreeTimes[SomeUnion]


def test_pep695_generic_alias_rename_three_times() -> None:
    """Adapted from: https://github.com/brentyi/tyro/issues/177"""

    @dataclass(frozen=True)
    class Config:
        arg: Annotated[SomeUnionRenamed, tyro.conf.arg(name="renamed_override")]

    assert tyro.cli(Config, args=["--renamed-override", "True"]) == Config(arg=True)


type RecursiveList[T] = T | list[RecursiveList[T]]


def test_pep695_recursive_types() -> None:
    """Adapted from: https://github.com/brentyi/tyro/issues/177"""

    @dataclass(frozen=True)
    class Config:
        arg: RecursiveList[str]

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(Config, args=["--arg", "True"])


def test_pep695_recursive_types_custom_constructor() -> None:
    """Adapted from: https://github.com/brentyi/tyro/issues/177"""

    @dataclass(frozen=True)
    class Config:
        arg: Annotated[RecursiveList[str], tyro.conf.arg(constructor=str)]

    assert tyro.cli(Config, args=["--arg", "True"]) == Config(arg="True")


type UnprefixedSubcommandPair[T1, T2, T3] = (
    Annotated[OmitArgPrefixes[T1], tyro.conf.subcommand(prefix_name=False)]  # type: ignore
    | Annotated[T2, tyro.conf.subcommand(prefix_name=False)]  # type: ignore
    | Annotated[T3, tyro.conf.subcommand(prefix_name=True)]
)
type IntContainer = Inner[int]
type IntContainerIntermediate = Inner[int]
type IntContainer2 = IntContainerIntermediate
type StrContainer = Inner[str]


def test_pep695_alias_subcommand() -> None:
    """Adapted from: https://github.com/brentyi/tyro/issues/177"""

    @dataclass(frozen=True)
    class Config:
        arg: UnprefixedSubcommandPair[IntContainer, IntContainer2, StrContainer]

    assert tyro.cli(Config, args=["int-container", "--a", "3", "--b", "5"]) == Config(
        Inner(3, 5)
    )
    assert tyro.cli(
        Config, args=["int-container2", "--arg.a", "3", "--arg.b", "5"]
    ) == Config(Inner(3, 5))
    assert tyro.cli(
        Config, args=["arg:str-container", "--arg.a", "3", "--arg.b", "5"]
    ) == Config(Inner("3", "5"))


type Int0 = int
Int1 = NewType("Int1", Int0)
type Int2 = Int1
Int3 = NewType("Int3", Int2)
type Int4 = Int3
Int5 = NewType("Int5", Int4)
type Int6 = Int5
Int7 = NewType("Int7", Int6)


def test_pep695_new_type_alias() -> None:
    def main(arg: list[Int7], /) -> Any:
        return arg

    assert tyro.cli(main, args=["1", "2"]) == [1, 2]


def test_generic_config():
    @dataclass(frozen=True)
    class Container[T]:
        a: Inner[T]

    assert tyro.cli(
        Container[bool],
        args="--a.a True --a.b False".split(" "),
        config=(tyro.conf.FlagConversionOff,),
    ) == Container(Inner(True, False))


def test_generic_config_subcommand():
    @dataclass(frozen=True)
    class Container[T]:
        a: T

    assert tyro.cli(
        Container[Container[bool] | Container[str]],
        args="a:container-bool --a.a True".split(" "),
        default=Container(Container(a="30")),
        config=(tyro.conf.FlagConversionOff,),
    ) == Container(Container(True))

    assert tyro.cli(
        Container[Container[bool] | Container[str]],
        args=[],
        default=Container(Container(a="30")),
        config=(tyro.conf.FlagConversionOff,),
    ) == Container(Container("30"))

    assert tyro.cli(
        Container[Container[bool] | Container[str]],
        args=[],
        default=Container(Container(a=False)),
        config=(tyro.conf.FlagConversionOff,),
    ) == Container(Container(False))


def test_generic_config_subcommand2():
    @dataclass(frozen=True)
    class Container[T]:
        a: tyro.conf.OmitSubcommandPrefixes[T]

    assert tyro.cli(
        Container[Container[bool] | Container[str]],
        args="container-bool --a True".split(" "),
    ) == Container(Container(True))


def test_generic_config_subcommand3():
    @dataclass(frozen=True)
    class Container[T]:
        a: T

    assert tyro.cli(
        Container[Container[bool] | Container[str]],
        args=[],
        default=Container(Container(a=True)),
        config=(tyro.conf.OmitSubcommandPrefixes,),
    ) == Container(Container(True))


def test_generic_config_subcommand4():
    @dataclass(frozen=True)
    class Container[T]:
        a: T

    assert tyro.cli(
        Container[Container[bool] | Container[Literal["1", "2"]]],
        args="container-literal-1-2 --a 2".split(" "),
        config=(tyro.conf.OmitSubcommandPrefixes,),
    ) == Container(Container("2"))

    assert tyro.cli(
        Container[Container[bool] | Container[Literal["1", "2"]]],
        args=[],
        default=Container(Container(a=True)),
        config=(tyro.conf.OmitSubcommandPrefixes,),
    ) == Container(Container(True))

    # This case is currently too hard for tyro's subcommand matcher.
    with pytest.raises(AssertionError):
        tyro.cli(
            Container[Container[bool] | Container[Literal["1", "2"]]],
            args=[],
            default=Container(Container(a="1")),
            config=(tyro.conf.OmitSubcommandPrefixes,),
        )



"""Counters

Repeatable 'counter' arguments can be specified via :data:`tyro.conf.UseCounterAction`.

Usage:

    python ./12_counters.py --help
    python ./12_counters.py --verbosity
    python ./12_counters.py --verbosity --verbosity
    python ./12_counters.py -vvv
"""

from typing_extensions import Annotated

import tyro
from tyro.conf import UseCounterAction


def main(
    verbosity: UseCounterAction[int],
    aliased_verbosity: Annotated[UseCounterAction[int], tyro.conf.arg(aliases=["-v"])],
) -> None:
    """Example showing how to use counter actions.

    Args:
        verbosity: Verbosity level.
        aliased_verbosity: Same as above, but can also be specified with -v, -vv, -vvv, etc.
    """
    print("Verbosity level:", verbosity)
    print("Verbosity level (aliased):", aliased_verbosity)


if __name__ == "__main__":
    tyro.cli(main)



from __future__ import annotations

import dataclasses
from typing import Any, Callable, Dict, Optional, Tuple, Union

from typing_extensions import get_args, get_origin

from tyro.constructors._struct_spec import UnsupportedStructTypeMessage

from . import _fields, _resolver, _singleton, _typing
from .conf import _confstruct


def match_subcommand(
    default: Any,
    subcommand_config_from_name: Dict[str, _confstruct._SubcommandConfig],
    subcommand_type_from_name: Dict[str, type],
) -> Optional[str]:
    """Given a subcommand mapping and a default, return which subcommand the default
    corresponds to.

    TOOD: this function is based on heuristics. While it should be robust to
    most real-world scenarios, there's room for improvement for generic types.
    """

    # Get default subcommand name: by default hash.
    default_hash = object.__hash__(default)
    for subcommand_name, conf in subcommand_config_from_name.items():
        if conf.default in _singleton.MISSING_AND_MISSING_NONPROP:
            continue
        if default_hash == object.__hash__(conf.default):
            return subcommand_name

    # Get default subcommand name: by default value.
    for subcommand_name, conf in subcommand_config_from_name.items():
        if conf.default in _singleton.MISSING_AND_MISSING_NONPROP:
            continue
        equal = default == conf.default
        if isinstance(equal, bool) and equal:
            return subcommand_name

    # Get default subcommand name: by concrete type tree.
    typetree = _TypeTree.make(type(default), default)
    for subcommand_name, conf in subcommand_config_from_name.items():
        if conf.default in _singleton.MISSING_AND_MISSING_NONPROP:
            continue
        if typetree == _TypeTree.make(type(conf.default), conf.default):
            return subcommand_name

    # Get default subcommand name: by annotated type tree.
    typetree_from_name = {
        subcommand_name: _TypeTree.make(subcommand_type, _singleton.MISSING_NONPROP)
        for subcommand_name, subcommand_type in subcommand_type_from_name.items()
    }
    for subcommand_name in subcommand_type_from_name.keys():
        # Equality check for type tree.
        if typetree == typetree_from_name[subcommand_name]:
            return subcommand_name
    for subcommand_name in subcommand_type_from_name.keys():
        # Leaky subtype check.
        if typetree.is_subtype_of(typetree_from_name[subcommand_name]):
            return subcommand_name

    # Failed. This should never happen, we'll raise an error outside of this function if
    # this is the case.
    return None  # pragma: no cover


@dataclasses.dataclass(frozen=True)
class _TypeTree:
    typ: Any
    children: Dict[str, _TypeTree]

    @staticmethod
    def make(
        typ: Union[Callable, _typing.TypeForm],
        default_instance: Any,
    ) -> _TypeTree:
        """From an object instance, return a data structure representing the types in the object."""

        typ_unwrap = _resolver.resolve_generic_types(typ)[0]

        field_list_out = _fields.field_list_from_type_or_callable(
            typ, default_instance=default_instance, support_single_arg_types=False
        )
        if isinstance(field_list_out, UnsupportedStructTypeMessage):
            return _TypeTree(typ_unwrap, {})

        typ, field_list = field_list_out
        return _TypeTree(
            typ_unwrap,
            {
                field.intern_name: _TypeTree.make(field.type_stripped, field.default)
                for field in field_list
            },
        )

    def is_subtype_of(self, supertype: _TypeTree) -> bool:
        """Best-effort subcommand matching."""

        # Generalize to unions.
        def _get_type_options(typ: _typing.TypeForm) -> Tuple[_typing.TypeForm, ...]:
            return get_args(typ) if get_origin(typ) is Union else (typ,)

        self_types = _get_type_options(self.typ)
        super_types = _get_type_options(supertype.typ)

        # Check against supertypes. TODO: the heuristics below could be more
        # principled. We should revisit.
        for self_type in self_types:
            self_type = _resolver.unwrap_annotated(self_type)
            ok = False
            for super_type in super_types:
                super_type = _resolver.unwrap_annotated(super_type)
                try:
                    if issubclass(self_type, super_type):
                        ok = True
                except TypeError:
                    pass
            if not ok:
                return False

        for child_name, child in self.children.items():
            if child_name not in supertype.children or not child.is_subtype_of(
                supertype.children[child_name]
            ):
                return False

        return True



# mypy: disable-error-code="call-overload,misc"
#
# Mypy errors from passing union types directly into tyro.cli() as Type[T]. We would
# benefit from TypeForm[T]: https://github.com/python/mypy/issues/9773
"""Tests adapted from https://github.com/brentyi/tyro/issues/89, which catches edge
cases when combining nested tuple types, renamed arguments, and subcommands.

Largely written by @wookayin.
"""

import dataclasses
from pathlib import Path
from typing import Annotated, Generic, Tuple, TypeVar

import tyro.conf

T = TypeVar("T")


@dataclasses.dataclass
class Checkout(Generic[T]):
    """Check out a branch."""

    branch: T


@dataclasses.dataclass
class Commit:
    """Commit something."""

    input: tyro.conf.Positional[Path]


@dataclasses.dataclass
class Arg:
    verbose: bool = True


def test_case1() -> None:
    o = tyro.cli(
        Checkout[str] | Commit,
        args=["commit", "./path.txt"],
    )
    assert o == Commit(input=Path("path.txt"))


def test_case2() -> None:
    arg, action = tyro.cli(
        Tuple[
            Arg,
            Annotated[
                Checkout[str] | Commit,
                tyro.conf.arg(name=""),
            ],
        ],
        args=["commit", "./path.txt"],
    )

    assert isinstance(arg, Arg)
    assert isinstance(action, Commit)
    assert action.input == Path("./path.txt")


def test_case3() -> None:
    o = tyro.cli(
        Tuple[
            Annotated[
                Arg,
                tyro.conf.arg(name=""),
            ],
            Annotated[
                Annotated[Checkout[str], tyro.conf.subcommand(name="checkout")]
                | Annotated[Commit, tyro.conf.subcommand(name="commit")],
                tyro.conf.arg(name=""),
            ],
        ],
        args=["commit", "./path.txt"],
    )
    assert o == (Arg(), Commit(Path("./path.txt")))


def test_case4() -> None:
    o = tyro.cli(
        Tuple[
            Annotated[
                Annotated[Checkout[str], tyro.conf.subcommand(name="checkout")]
                | Annotated[Commit, tyro.conf.subcommand(name="commit")],
                tyro.conf.arg(name=""),
            ]
        ],
        args=["commit", "./path.txt"],
    )
    assert o == (Commit(Path("./path.txt")),)


def test_case5() -> None:
    assert tyro.cli(
        tyro.conf.OmitArgPrefixes[
            Tuple[
                Annotated[
                    Checkout[str],
                    tyro.conf.subcommand(prefix_name=False),
                ]
                | Annotated[Commit, tyro.conf.subcommand(prefix_name=False)],
                Arg,
            ]
        ],
        args=["--no-verbose", "checkout-str", "--branch", "branch"],
    ) == (Checkout("branch"), Arg(False))



"""Rules for taking high-level field definitions and lowering them into inputs for
argparse's `add_argument()`."""

from __future__ import annotations

import dataclasses
import json
import shlex
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import rich.markup
import shtab
from typing_extensions import get_origin

from . import _argparse as argparse
from . import _fields, _singleton, _strings
from .conf import _markers
from .constructors import (
    ConstructorRegistry,
    PrimitiveTypeInfo,
    UnsupportedTypeAnnotationError,
)

if TYPE_CHECKING:
    cached_property = property
else:
    try:
        # Python >=3.8.
        from functools import cached_property
    except ImportError:
        # Python 3.7.
        from backports.cached_property import cached_property  # type: ignore


_T = TypeVar("_T")


class BooleanOptionalAction(argparse.Action):
    """Adapted from https://github.com/python/cpython/pull/27672"""

    def __init__(
        self,
        option_strings: Sequence[str],
        dest: str,
        default: _T | str | None = None,
        type: Callable[[str], _T] | argparse.FileType | None = None,
        choices: Iterable[_T] | None = None,
        required: bool = False,
        help: str | None = None,
        metavar: str | tuple[str, ...] | None = None,
    ) -> None:
        _option_strings = []
        self._no_strings = set()
        for option_string in option_strings:
            _option_strings.append(option_string)

            if option_string.startswith("--"):
                if "." not in option_string:
                    option_string = (
                        "--no" + _strings.get_delimeter() + option_string[2:]
                    )
                else:
                    # Loose heuristic for where to add the no-/no_ prefix.
                    left, _, right = option_string.rpartition(".")
                    option_string = left + ".no" + _strings.get_delimeter() + right
                self._no_strings.add(option_string)

                _option_strings.append(option_string)

        super().__init__(
            option_strings=_option_strings,
            dest=dest,
            nargs=0,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        if option_string in self.option_strings:
            assert option_string is not None
            setattr(namespace, self.dest, option_string not in self._no_strings)

    # Typically only supported in Python 3.10, but we backport some functionality in
    # _argparse_formatters.py
    def format_usage(self):
        return " | ".join(self.option_strings)


@dataclasses.dataclass(frozen=True)
class ArgumentDefinition:
    """Structure containing everything needed to define an argument."""

    intern_prefix: str  # True prefix. (eg for the argument's dest field)
    extern_prefix: str  # User-facing prefix.
    subcommand_prefix: str  # Prefix for nesting.
    field: _fields.FieldDefinition

    def add_argument(
        self, parser: Union[argparse.ArgumentParser, argparse._ArgumentGroup]
    ) -> None:
        """Add a defined argument to a parser."""

        # Get keyword arguments, with None values removed.
        kwargs = dict(self.lowered.__dict__)  # type: ignore
        kwargs.pop("instance_from_str")
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        name_or_flag = kwargs.pop("name_or_flag")
        if len(name_or_flag) == 0:
            name_or_flag = _strings.dummy_field_name

        # We're actually going to skip the default field: if an argument is unset, the
        # MISSING value will be detected in _calling.py and the field default will
        # directly be used. This helps reduce the likelihood of issues with converting
        # the field default to a string format, then back to the desired type.
        action = kwargs.get("action", None)
        if action not in {"append", "count"}:
            kwargs["default"] = _singleton.MISSING_NONPROP
        elif action in {BooleanOptionalAction, "count"}:
            pass
        else:
            kwargs["default"] = []

        # Apply overrides in our arg configuration object.
        # The `name` field is applied when the field object is instantiated!
        if self.field.argconf.metavar is not None:
            kwargs["metavar"] = self.field.argconf.metavar

        # Add argument, with aliases if available.
        if self.field.argconf.aliases is not None and not self.field.is_positional():
            arg = parser.add_argument(
                name_or_flag, *self.field.argconf.aliases, **kwargs
            )
        else:
            if self.field.argconf.aliases is not None:
                import warnings

                warnings.warn(
                    f"Aliases were specified, but {name_or_flag} is positional. Aliases will be ignored."
                )
            arg = parser.add_argument(name_or_flag, **kwargs)

        # Do our best to tab complete paths.
        # There will be false positives here, but if choices is unset they should be
        # harmless.
        if "choices" not in kwargs:
            name_suggests_dir = (
                # The conditions are intended to be conservative; if a directory path is
                # registered as a normal file one that's OK, the reverse on the other
                # hand will be overly restrictive.
                self.field.intern_name.endswith("_dir")
                or self.field.intern_name.endswith("_directory")
                or self.field.intern_name.endswith("_folder")
            )
            name_suggests_path = (
                self.field.intern_name.endswith("_file")
                or self.field.intern_name.endswith("_path")
                or self.field.intern_name.endswith("_filename")
                or name_suggests_dir
            )
            complete_as_path = (
                # Catch types like Path, List[Path], Tuple[Path, ...] etc.
                "Path" in str(self.field.type_stripped)
                # For string types, we require more evidence.
                or ("str" in str(self.field.type_stripped) and name_suggests_path)
            )
            if complete_as_path:
                arg.complete = shtab.DIRECTORY if name_suggests_dir else shtab.FILE  # type: ignore

    @cached_property
    def lowered(self) -> LoweredArgumentDefinition:
        """Lowered argument definition, generated by applying a sequence of rules."""
        # Each rule will mutate the lowered object. This is (unfortunately)
        # much faster than a functional approach.
        lowered = LoweredArgumentDefinition()
        _rule_handle_boolean_flags(self, lowered)
        _rule_apply_primitive_specs(self, lowered)
        _rule_counters(self, lowered)
        _rule_generate_helptext(self, lowered)
        _rule_set_name_or_flag_and_dest(self, lowered)
        _rule_positional_special_handling(self, lowered)
        return lowered


@dataclasses.dataclass
class LoweredArgumentDefinition:
    """Contains fields meant to be passed directly into argparse."""

    # Action that is called on parsed arguments. This handles conversions from strings
    # to our desired types.
    #
    # The main reason we use this instead of the standard 'type' argument is to enable
    # mixed-type tuples.
    instance_from_str: Optional[Callable] = None

    def is_fixed(self) -> bool:
        """If the instantiator is set to `None`, even after all argument
        transformations, it means that we don't have a valid instantiator for an
        argument. We then mark the argument as 'fixed', with a value always equal to the
        field default."""
        return self.instance_from_str is None

    # From here on out, all fields correspond 1:1 to inputs to argparse's
    # add_argument() method.
    name_or_flag: str = ""
    default: Optional[Any] = None
    dest: Optional[str] = None
    required: Optional[bool] = None
    action: Optional[Any] = None
    nargs: Optional[Union[int, str]] = None
    choices: Optional[Tuple[str, ...]] = None
    # Note: unlike in vanilla argparse, our metavar is always a string. We handle
    # sequences, multiple arguments, etc, manually.
    metavar: Optional[str] = None
    help: Optional[str] = None


def _rule_handle_boolean_flags(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> None:
    if arg.field.type_stripped is not bool:
        return

    if (
        arg.field.default in _singleton.MISSING_AND_MISSING_NONPROP
        or arg.field.is_positional()
        or _markers.FlagConversionOff in arg.field.markers
        or _markers.Fixed in arg.field.markers
    ):
        # Treat bools as a normal parameter.
        return
    elif arg.field.default in (True, False):
        # Default `False` => --flag passed in flips to `True`.
        lowered.action = BooleanOptionalAction
        lowered.instance_from_str = (
            lambda x: x
        )  # argparse will directly give us a bool!
        lowered.default = arg.field.default
        return

    assert False, (
        f"Expected a boolean as a default for {arg.field.intern_name}, but got"
        f" {arg.field.default}."
    )


def _rule_apply_primitive_specs(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> None:
    """The bulkiest bit: recursively analyze the type annotation and use it to determine
    how to instantiate it given some string from the commandline.

    Important: as far as argparse is concerned, all inputs are strings.

    Conversions from strings to our desired types happen in the instantiator; this is a
    bit more flexible, and lets us handle more complex types like enums and multi-type
    tuples."""

    if _markers.Fixed in arg.field.markers:
        lowered.instance_from_str = None
        lowered.metavar = "{fixed}"
        lowered.required = False
        lowered.default = _singleton.MISSING
        return
    if lowered.instance_from_str is not None:
        return

    try:
        spec = ConstructorRegistry._get_active_registry().get_primitive_spec(
            PrimitiveTypeInfo.make(
                cast(type, arg.field.type),
                arg.field.markers,
            )
        )
    except UnsupportedTypeAnnotationError as e:
        if arg.field.default in _singleton.MISSING_AND_MISSING_NONPROP:
            field_name = _strings.make_field_name(
                [arg.extern_prefix, arg.field.intern_name]
            )
            if field_name != "":
                raise UnsupportedTypeAnnotationError(
                    f"Unsupported type annotation for the field {field_name}; "
                    f"{e.args[0]} "
                    "To suppress this error, assign the field either a default value or a different type."
                ) from e
            else:
                # If the field name is empty, it means we're raising an error
                # for the direct input to `tyro.cli()`. We don't need to write
                # out which specific field we're complaining about.
                raise e
        else:
            # For fields with a default, we'll get by even if there's no instantiator
            # available.
            lowered.metavar = "{fixed}"
            lowered.required = False
            lowered.default = _singleton.MISSING
            return

    # Mark lowered as required if a default is missing.
    if (
        arg.field.default in _singleton.MISSING_AND_MISSING_NONPROP
        and _markers._OPTIONAL_GROUP not in arg.field.markers
    ):
        lowered.default = None
        lowered.required = True
    elif (
        arg.field.default is not _singleton.EXCLUDE_FROM_CALL
        and arg.field.default not in _singleton.MISSING_AND_MISSING_NONPROP
    ):
        # Set default.
        lowered.default = spec.str_from_instance(arg.field.default)
    else:
        lowered.default = arg.field.default

    if spec._action == "append":

        def append_instantiator(x: list[list[str]]) -> Any:
            """Handle UseAppendAction effects."""
            # We'll assume that the type is annotated as Dict[...], Tuple[...], List[...], etc.
            container_type = get_origin(arg.field.type_stripped)
            if container_type is None:
                # Raw annotation, like `UseAppendAction[list]`. It's unlikely
                # that a user would use this but we can handle it.
                container_type = arg.field.type_stripped

            # Instantiate initial output.
            out = (
                arg.field.default
                if arg.field.default not in _singleton.MISSING_AND_MISSING_NONPROP
                else None
            )
            if out is None:
                out = {} if container_type is dict else []
            elif isinstance(out, dict):
                out = out.copy()
            else:
                # All sequence types will be lists for now to make sure we can
                # append to them.
                out = list(out)

            # Get + merge parts.
            parts = [spec.instance_from_str(arg_list) for arg_list in x]
            for part in parts:
                if isinstance(out, dict):
                    out.update(part)
                else:
                    out.append(part)

            # Return output with correct type.
            if isinstance(out, dict):
                return out
            else:
                return container_type(out)

        lowered.instance_from_str = append_instantiator
        lowered.default = None
        lowered.choices = spec.choices
        lowered.nargs = spec.nargs
        lowered.metavar = spec.metavar
        lowered.action = spec._action
        lowered.required = False
        return
    else:
        lowered.instance_from_str = spec.instance_from_str
        lowered.choices = spec.choices
        lowered.nargs = spec.nargs
        lowered.metavar = spec.metavar
        lowered.action = spec._action
        return


# This can be turned off when we don't want rich-based formatting. (notably for
# completion scripts)
#
# TODO: the global state here is gross. Should be revisited.
USE_RICH = True


# TODO: this function is also called outside of _arguments.py. Should be revisited.
def _rich_tag_if_enabled(x: str, tag: str) -> str:
    x = rich.markup.escape(_strings.strip_ansi_sequences(x))
    return x if not USE_RICH else f"[{tag}]{x}[/{tag}]"


def _rule_counters(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> None:
    """Handle counters, like -vvv for level-3 verbosity."""
    if (
        _markers.UseCounterAction in arg.field.markers
        and arg.field.type_stripped is int
        and not arg.field.is_positional()
    ):
        lowered.metavar = None
        lowered.nargs = None
        lowered.action = "count"
        lowered.default = 0
        lowered.required = False
        lowered.instance_from_str = (
            lambda x: x
        )  # argparse will directly give us an int!
        return


def _rule_generate_helptext(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> None:
    """Generate helptext from docstring, argument name, default values."""

    # If the suppress marker is attached, hide the argument.
    if _markers.Suppress in arg.field.markers or (
        _markers.SuppressFixed in arg.field.markers and lowered.is_fixed()
    ):
        lowered.help = argparse.SUPPRESS
        return

    help_parts = []

    primary_help = arg.field.helptext

    if primary_help is None and _markers.Positional in arg.field.markers:
        primary_help = _strings.make_field_name(
            [arg.extern_prefix, arg.field.intern_name]
        )

    if primary_help is not None:
        help_parts.append(_rich_tag_if_enabled(primary_help, "helptext"))

    if not lowered.required:
        # Get the default value.
        # Note: lowered.default is the stringified version!
        default = lowered.default
        if lowered.is_fixed() or lowered.action == "append":
            # Cases where we'll be missing the lowered default. Use field default instead.
            assert default in _singleton.MISSING_AND_MISSING_NONPROP or default is None
            default = arg.field.default

        # Get the default value label.
        if arg.field.argconf.constructor_factory is not None:
            default_label = (
                str(default)
                if arg.field.type_stripped is not json.loads
                else json.dumps(arg.field.default)
            )
        elif type(default) in (tuple, list, set):
            # For tuple types, we might have default as (0, 1, 2, 3).
            # For list types, we might have default as [0, 1, 2, 3].
            # For set types, we might have default as {0, 1, 2, 3}.
            #
            # In all cases, we want to display (default: 0 1 2 3), for consistency with
            # the format that argparse expects when we set nargs.
            assert default is not None
            default_label = " ".join(map(shlex.quote, map(str, default)))
        else:
            default_label = str(default)

        # Suffix helptext with some behavior hint, such as the default value of the argument.
        help_behavior_hint = arg.field.argconf.help_behavior_hint
        if help_behavior_hint is not None:
            behavior_hint = (
                help_behavior_hint(default_label)
                if callable(help_behavior_hint)
                else help_behavior_hint
            )
        elif lowered.instance_from_str is None:
            # Intentionally not quoted via shlex, since this can't actually be passed
            # in via the commandline.
            behavior_hint = f"(fixed to: {default_label})"
        elif lowered.action == "count":
            # Repeatable argument.
            behavior_hint = "(repeatable)"
        elif lowered.action == "append" and (
            default in _singleton.MISSING_AND_MISSING_NONPROP
            or len(cast(tuple, default)) == 0
        ):
            behavior_hint = "(repeatable)"
        elif lowered.action == "append" and len(cast(tuple, default)) > 0:
            assert default is not None  # Just for type checker.
            behavior_hint = f"(repeatable, appends to: {default_label})"
        elif arg.field.default is _singleton.EXCLUDE_FROM_CALL:
            # ^important to use arg.field.default and not the stringified default variable.
            behavior_hint = "(unset by default)"
        elif (
            _markers._OPTIONAL_GROUP in arg.field.markers
            and default in _singleton.MISSING_AND_MISSING_NONPROP
        ):
            # Argument in an optional group, but with no default. This is typically used
            # when general (non-argument, non-dataclass) object arguments are given a
            # default, or when we use `tyro.conf.arg(constructor=...)`.
            #
            # There are some usage details that aren't communicated right now in the
            # helptext. For example: all arguments within an optional group without a
            # default should be passed in or none at all.
            behavior_hint = "(optional)"
        elif _markers._OPTIONAL_GROUP in arg.field.markers:
            # Argument in an optional group, but which also has a default.
            behavior_hint = f"(default if used: {default_label})"
        else:
            behavior_hint = f"(default: {default_label})"

        help_parts.append(_rich_tag_if_enabled(behavior_hint, "helptext_default"))
    else:
        help_parts.append(_rich_tag_if_enabled("(required)", "helptext_required"))

    # The percent symbol needs some extra handling in argparse.
    # https://stackoverflow.com/questions/21168120/python-argparse-errors-with-in-help-string
    lowered.help = " ".join([p for p in help_parts if len(p) > 0]).replace("%", "%%")
    return


def _rule_set_name_or_flag_and_dest(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> None:
    if lowered.help is argparse.SUPPRESS:
        # Use standard name for suppressed arguments.
        # Relevant: https://github.com/brentyi/tyro/issues/170
        name_or_flag = _strings.make_field_name(
            [arg.extern_prefix, arg.field.extern_name]
        )
    elif (
        arg.field.argconf.prefix_name is False
        or _markers.OmitArgPrefixes in arg.field.markers
    ):
        # Strip prefixes when the argument is suppressed.
        # Still need to call make_field_name() because it converts underscores
        # to hyphens, etc.
        name_or_flag = _strings.make_field_name([arg.field.extern_name])
    elif (
        _markers.OmitSubcommandPrefixes in arg.field.markers
        and arg.subcommand_prefix != ""
    ):
        # Strip subcommand prefixes, but keep following
        # prefixes.`extern_prefix` can start with the prefix corresponding to
        # the parent subcommand, but end with other prefixes correspondeding to
        # nested structures within the subcommand.
        name_or_flag = _strings.make_field_name(
            [arg.extern_prefix, arg.field.extern_name]
        )
        strip_prefix = arg.subcommand_prefix + "."
        assert name_or_flag.startswith(strip_prefix), name_or_flag
        name_or_flag = name_or_flag[len(strip_prefix) :]
    else:
        # Standard prefixed name.
        name_or_flag = _strings.make_field_name(
            [arg.extern_prefix, arg.field.extern_name]
        )

    # Prefix keyword arguments with --.
    if not arg.field.is_positional():
        name_or_flag = "--" + name_or_flag

    lowered.name_or_flag = name_or_flag
    lowered.dest = _strings.make_field_name([arg.intern_prefix, arg.field.intern_name])


def _rule_positional_special_handling(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> None:
    if not arg.field.is_positional():
        return None

    metavar = lowered.metavar
    if lowered.required:
        nargs = lowered.nargs
    else:
        if metavar is not None:
            metavar = "[" + metavar + "]"
        if lowered.nargs == 1:
            # Optional positional arguments. This needs to be special-cased in
            # _calling.py.
            nargs = "?"
        else:
            # If lowered.nargs is either + or an int.
            nargs = "*"

    lowered.name_or_flag = _strings.make_field_name(
        [arg.intern_prefix, arg.field.intern_name]
    )
    lowered.dest = None
    lowered.required = None  # Can't be passed in for positionals.
    lowered.metavar = metavar
    lowered.nargs = nargs
    return



import dataclasses

from helptext_utils import get_helptext_with_checks

import tyro


def test_flag_default_false() -> None:
    """Test for argparse.BooleanOptionalAction-style usage."""

    @dataclasses.dataclass
    class A:
        x: bool

    assert tyro.cli(
        A,
        args=["--x"],
        default=A(False),
    ) == A(True)

    assert tyro.cli(
        A,
        args=["--no-x"],
        default=A(False),
    ) == A(False)

    assert tyro.cli(
        A,
        args=[],
        default=A(False),
    ) == A(False)

    # Type ignore can be removed once TypeForm lands.
    # https://discuss.python.org/t/typeform-spelling-for-a-type-annotation-object-at-runtime/51435
    assert tyro.cli(
        tyro.conf.FlagConversionOff[A],
        args=["--x", "True"],
        default=A(False),  # type: ignore
    ) == A(True)


def test_flag_default_true() -> None:
    """Test for argparse.BooleanOptionalAction-style usage."""

    @dataclasses.dataclass
    class A:
        x: bool

    assert tyro.cli(
        A,
        args=["--x"],
        default=A(True),
    ) == A(True)

    assert tyro.cli(
        A,
        args=["--no-x"],
        default=A(True),
    ) == A(False)

    assert tyro.cli(
        A,
        args=[],
        default=A(True),
    ) == A(True)

    # Type ignore can be removed once TypeForm lands.
    # https://discuss.python.org/t/typeform-spelling-for-a-type-annotation-object-at-runtime/51435
    assert tyro.cli(
        tyro.conf.FlagConversionOff[A],
        args=["--x", "True"],
        default=A(False),  # type: ignore
    ) == A(True)


def test_flag_default_true_helptext() -> None:
    """Test for argparse.BooleanOptionalAction-style usage."""

    @dataclasses.dataclass
    class A:
        x: bool = True

    assert "(default: True)" in get_helptext_with_checks(A)
    assert "(default: False)" not in get_helptext_with_checks(A)
    assert "(default: None)" not in get_helptext_with_checks(A)



# mypy: ignore-errors
"""Tests adapted from https://github.com/brentyi/tyro/issues/89, which catches edge
cases when combining nested tuple types, renamed arguments, and subcommands.

Largely written by @wookayin.
"""

import dataclasses
from pathlib import Path
from typing import Annotated, cast

import tyro.conf


@dataclasses.dataclass
class Checkout[T]:
    """Check out a branch."""

    branch: T


@dataclasses.dataclass
class Commit:
    """Commit something."""

    input: tyro.conf.Positional[Path]


@dataclasses.dataclass
class Arg:
    verbose: bool = True


def test_case1() -> None:
    o = tyro.cli(
        Checkout[str] | Commit,
        args=["commit", "./path.txt"],
    )
    assert o == Commit(input=Path("path.txt"))


def test_case2() -> None:
    arg, action = tyro.cli(
        tuple[
            Arg,
            Annotated[
                Checkout[str] | Commit,
                tyro.conf.arg(name=""),
            ],
        ],
        args=["commit", "./path.txt"],
    )

    assert isinstance(arg, Arg)
    assert isinstance(action, Commit)
    assert action.input == Path("./path.txt")


def test_case3() -> None:
    o = tyro.cli(
        tuple[
            Annotated[
                Arg,
                tyro.conf.arg(name=""),
            ],
            Annotated[
                Annotated[Checkout[str], tyro.conf.subcommand(name="checkout")]
                | Annotated[Commit, tyro.conf.subcommand(name="commit")],
                tyro.conf.arg(name=""),
            ],
        ],
        args=["commit", "./path.txt"],
    )
    assert o == (Arg(), Commit(Path("./path.txt")))


def test_case4() -> None:
    o = tyro.cli(
        tuple[
            Annotated[
                Annotated[Checkout[str], tyro.conf.subcommand(name="checkout")]
                | Annotated[Commit, tyro.conf.subcommand(name="commit")],
                tyro.conf.arg(name=""),
            ]
        ],
        args=["commit", "./path.txt"],
    )
    assert o == (Commit(Path("./path.txt")),)


def test_case5() -> None:
    assert tyro.cli(
        tyro.conf.OmitArgPrefixes[
            tuple[
                Annotated[
                    Checkout[str],
                    tyro.conf.subcommand(prefix_name=False),
                ]
                | Annotated[Commit, tyro.conf.subcommand(prefix_name=False)],
                Arg,
            ]
        ],
        args=["--no-verbose", "checkout-str", "--branch", "branch"],
    ) == (Checkout("branch"), Arg(False))


# New tests using PEP 695 type aliases:

type CheckoutAlias[T] = Annotated[Checkout[T], tyro.conf.subcommand(name="checkout")]
type CommitAlias = Annotated[Commit, tyro.conf.subcommand(name="commit")]
type ActionUnion = CheckoutAlias[str] | CommitAlias
type RenamedArg = Annotated[Arg, tyro.conf.arg(name="global")]


def test_case6() -> None:
    o = tyro.cli(
        tuple[
            RenamedArg,
            Annotated[ActionUnion, tyro.conf.arg(name="")],
        ],
        args=["--global.no-verbose", "commit", "./path.txt"],
    )
    assert o == (Arg(verbose=False), Commit(Path("./path.txt")))


# https://github.com/microsoft/pyright/issues/9261
type PositionalPath = tyro.conf.Positional[Path]  # type: ignore


def test_case7() -> None:
    @dataclasses.dataclass
    class NewCommit:
        """Commit something with a message."""

        input: PositionalPath
        message: str = "Default commit message"

    o = tyro.cli(
        cast(
            type,
            CheckoutAlias[str]
            | Annotated[NewCommit, tyro.conf.subcommand(name="commit")],
        ),
        args=["commit", "./path.txt", "--message", "New commit"],
    )
    assert o == NewCommit(input=Path("./path.txt"), message="New commit")


type VerboseArg = Annotated[
    bool, tyro.conf.arg(name="verbose", help="Enable verbose output")
]


def test_case8() -> None:
    @dataclasses.dataclass
    class ConfigWithVerbose:
        action: ActionUnion
        verbose: VerboseArg = False

    o = tyro.cli(
        ConfigWithVerbose,
        args=["--verbose", "action:checkout", "--action.branch", "main"],
    )
    assert o == ConfigWithVerbose(verbose=True, action=Checkout(branch="main"))



"""Enums

In addition to literals, enums can also be used to provide a fixed set of
choices.

Usage:

    python ./06_enums.py --help
    python ./06_enums.py --color RED
    python ./06_enums.py --color BLUE --opacity 0.75
"""

import enum
from dataclasses import dataclass
from pprint import pprint

import tyro


class Color(enum.Enum):
    RED = enum.auto()
    BLUE = enum.auto()


@dataclass
class Config:
    color: Color = Color.RED
    """Color argument."""

    opacity: float = 0.5
    """Opacity argument."""


if __name__ == "__main__":
    config = tyro.cli(Config)
    pprint(config)
